#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.

"""Functions to work with templates."""

import functools
import math
import numbers
import os
import pathlib
import warnings

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.colors as mcl
import networkx as nx
import seaborn as sns
import trimesh as tm

from matplotlib.lines import Line2D
from scipy.spatial.distance import pdist

from collections import namedtuple
from typing import List, Union, Optional
from typing_extensions import Literal

from .. import config, core, utils

from . import factory
from .base import TransformSequence, BaseTransform, AliasTransform
from .xfm_funcs import mirror, xform

# Catch some stupid warning about installing python-Levenshtein
with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    import fuzzywuzzy as fw
    import fuzzywuzzy.process

__all__ = ['xform_brain', 'mirror_brain', 'symmetrize_brain']

logger = config.get_logger(__name__)

# Defines entry the registry needs to register a transform
transform_reg = namedtuple('Transform',
                           ['source', 'target', 'transform', 'type',
                            'invertible', 'weight'])

# Check for environment variable pointing to registries
_OS_TRANSPATHS = os.environ.get('NAVIS_TRANSFORMS', '')
try:
    _OS_TRANSPATHS = [i for i in _OS_TRANSPATHS.split(';') if len(i) > 0]
except BaseException:
    logger.error('Error parsing the `NAVIS_TRANSFORMS` environment variable')
    _OS_TRANSPATHS = []


class TemplateRegistry:
    """Tracks template brains, available transforms and produces bridging sequences.

    Parameters
    ----------
    scan_paths :    bool
                    If True will scan paths on initialization.

    """
    def __init__(self, scan_paths: bool = True):
        # Paths to scan for transforms
        self._transpaths = _OS_TRANSPATHS.copy()
        # Transforms
        self._transforms = []
        # Template brains
        self._templates = []

        if scan_paths:
            self.scan_paths()

    def __contains__(self, other) -> bool:
        """Check if transform is in registry.

        Parameters
        ----------
        other :     transform, filepath, tuple
                    Either a transform (e.g. CMTKtransform), a filepath (e.g.
                    to a .list file) or a tuple of `(source, target, transform)`
                    where `transform` can be a transform or a filepath.

        """
        if isinstance(other, (tuple, list)):
            return any([t == other for t in self.transforms])
        else:
            return other in [t.transform for t in self.transforms]

    def __len__(self) -> int:
        return len(self.transforms)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f'TemplateRegistry with {len(self)} transforms'

    @property
    def transpaths(self) -> list:
        """Paths searched for transforms.

        Use `.scan_paths` to trigger a scan. Use `.register_path` to add
        more path(s).
        """
        return self._transpaths

    @property
    def templates(self) -> list:
        """Registered template (brains)."""
        return self._templates

    @property
    def transforms(self) -> list:
        """Registered transforms (bridging + mirror)."""
        return self._transforms

    @property
    def bridges(self) -> list:
        """Registered bridging transforms."""
        return [t for t in self.transforms if t.type == 'bridging']

    @property
    def mirrors(self) ->list:
        """Registered mirror transforms."""
        return [t for t in self.transforms if t.type == 'mirror']

    def clear_caches(self):
        """Clear caches of all cached functions."""
        self.bridging_graph.cache_clear()
        self.shortest_bridging_seq.cache_clear()

    def summary(self) -> pd.DataFrame:
        """Generate summary of available transforms."""
        return pd.DataFrame(self.transforms)

    def register_path(self, paths: str, trigger_scan: bool = True):
        """Register path(s) to scan for transforms.

        Parameters
        ----------
        paths :         str | list thereof
                        Paths (or list thereof) to scans for transforms. This
                        is not permanent. For permanent additions set path(s)
                        via the `NAVIS_TRANSFORMS` environment variable.
        trigger_scan :  bool
                        If True, a re-scan of all paths will be triggered.

        """
        paths = utils.make_iterable(paths)

        for p in paths:
            # Try not to duplicate paths
            if p not in self.transpaths:
                self._transpaths.append(p)

        if trigger_scan:
            self.scan_paths()

    def register_templatebrain(self, template: 'TemplateBrain',
                               skip_existing=True):
        """Register a template brain.

        This is used, for example, by navis.mirror_brain.

        Parameters
        ----------
        template :      TemplateBrain
                        TemplateBrain to register.
        skip_existing : bool
                        If True, will skip existing template brains.

        """
        utils.eval_param(template,
                         name='template',
                         allowed_types=(TemplateBrain, ))

        if template not in self._templates or not skip_existing:
            self._templates.append(template)

    def register_transform(self, transform: BaseTransform, source: str,
                           target: str, transform_type: str,
                           skip_existing: bool = True,
                           weight: int = 1):
        """Register a transform.

        Parameters
        ----------
        transform :         subclass of BaseTransform | TransformSequence
                            A transform (AffineTransform, CMTKtransform, etc.)
                            or a TransformSequence.
        source :            str
                            Source for forward transform.
        target :            str
                            Target for forward transform. Ignored for mirror
                            transforms.
        transform_type :    "bridging" | "mirror"
                            Type of transform.
        skip_existing :     bool
                            If True will skip if transform is already in registry.
        weight :            int
                            Giving a transform a lower weight will make it
                            preferable when plotting bridging sequences.

        See Also
        --------
        register_transformfile
                            If you want to register a file instead of an
                            already constructed transform.

        """
        assert transform_type in ('bridging', 'mirror')
        assert isinstance(transform, (BaseTransform, TransformSequence))

        # Translate into edge
        edge = transform_reg(source=source, target=target, transform=transform,
                             type=transform_type,
                             invertible=hasattr(transform, '__neg__'),
                             weight=weight)

        # Don't add if already exists
        if not skip_existing or edge not in self:
            self.transforms.append(edge)

        # Clear cached functions
        self.clear_caches()

    def register_transformfile(self, path: str, **kwargs):
        """Parse and register a transform file.

        File/Directory name must follow the a `{TARGET}_{SOURCE}.{ext}`
        convention (e.g. `JRC2013_FCWB.list`).

        Parameters
        ----------
        path :          str
                        Path to transform.
        **kwargs
                        Keyword arguments are passed to the constructor of the
                        Transform (e.g. CMTKtransform for `.list` directory).

        See Also
        --------
        register_transform
                        If you want to register an already constructed transform
                        instead of a transform file that still needs to be
                        parsed.

        """
        assert isinstance(path, (str, pathlib.Path))

        path = pathlib.Path(path).expanduser()

        if not path.is_dir() and not path.is_file():
            raise ValueError(f'File/directory "{path}" does not exist')

        # Parse properties
        try:
            if 'mirror' in path.name or 'imgflip' in path.name:
                transform_type = 'mirror'
                source = path.name.split('_')[0]
                target = None
            else:
                transform_type = 'bridging'
                target = path.name.split('_')[0]
                source = path.name.split('_')[1].split('.')[0]

            # Initialize the transform
            transform = factory.factory_methods[path.suffix](path, **kwargs)

            self.register_transform(transform=transform,
                                    source=source,
                                    target=target,
                                    transform_type=transform_type)
        except BaseException as e:
            logger.error(f'Error registering {path} as transform: {str(e)}')

    def scan_paths(self, extra_paths: List[str] = None):
        """Scan registered paths for transforms and add to registry.

        Will skip transforms that already exist in this registry.

        Parameters
        ----------
        extra_paths :   list of str
                        Any Extra paths to search.

        """
        search_paths = self.transpaths

        if isinstance(extra_paths, str):
            extra_paths = [i for i in extra_paths.split(';') if len(i) > 0]
            search_paths = np.append(search_paths, extra_paths)

        for path in search_paths:
            path = pathlib.Path(path).expanduser()
            # Skip if path does not exist
            if not path.is_dir():
                continue

            # Go over the file extensions we can work with (.h5, .list, .json)
            # These file extensions are registered in the
            # `navis.transforms.factory` module
            for ext in factory.factory_methods:
                for hit in path.rglob(f'*{ext}'):
                    if hit.is_dir() or hit.is_file():
                        # Register this file
                        self.register_transformfile(hit)

        # Clear cached functions
        self.clear_caches()

    @functools.lru_cache()
    def bridging_graph(self,
                       reciprocal: Union[Literal[False], int, float] = True) -> nx.DiGraph:
        """Generate networkx Graph describing the bridging paths.

        Parameters
        ----------
        reciprocal :    bool | float
                        If True or float, will add forward and inverse edges for
                        transforms that are invertible. If float, the inverse
                        edges' weights will be scaled by that factor.

        Returns
        -------
        networkx.MultiDiGraph

        """
        # Drop mirror transforms
        bridge = [t for t in self.transforms if t.type == 'bridging']
        bridge_inv = [t for t in bridge if t.invertible]

        # Generate graph
        # Note we are using MultiDi graph here because we might
        # have multiple edges between nodes. For example, there
        # is a JFRC2013DS_JFRC2013 and a JFRC2013_JFRC2013DS
        # bridging registration. If we include the inverse, there
        # will be two edges connecting JFRC2013DS and JFRC2013 in
        # both directions
        G = nx.MultiDiGraph()
        edges = [(t.source, t.target,
                  {'transform': t.transform,
                   'type': type(t.transform).__name__,
                   'weight': t.weight}) for t in bridge]

        if reciprocal:
            if isinstance(reciprocal, numbers.Number):
                rv_edges = [(t.target, t.source,
                             {'transform': -t.transform,  # note inverse transform!
                              'type': str(type(t.transform)).split('.')[-1],
                              'weight': t.weight * reciprocal}) for t in bridge_inv]
            else:
                rv_edges = [(t.target, t.source,
                             {'transform': -t.transform,  # note inverse transform!
                              'type': str(type(t.transform)).split('.')[-1],
                              'weight': t.weight}) for t in bridge_inv]
            edges += rv_edges

        G.add_edges_from(edges)

        return G

    def find_bridging_path(self, source: str,
                           target: str,
                           via: Optional[str] = None,
                           avoid: Optional[str] = None,
                           reciprocal=True) -> tuple:
        """Find bridging path from source to target.

        Parameters
        ----------
        source :        str
                        Source from which to transform to `target`.
        target :        str
                        Target to which to transform to.
        via :           str | list thereof, optional
                        Force specific intermediate template(s).
        avoid :         str | list thereof, optional
                        Avoid going through specific intermediate template(s).
        reciprocal :    bool | float
                        If True or float, will add forward and inverse edges for
                        transforms that are invertible. If float, the inverse
                        edges' weights will be scaled by that factor.

        Returns
        -------
        path :          list
                        Path from source to target: [source, ..., target]
        transforms :    list
                        Transforms as [[path_to_transform, inverse], ...]

        """
        # Generate (or get cached) bridging graph
        G = self.bridging_graph(reciprocal=reciprocal)

        if len(G) == 0:
            raise ValueError('No bridging registrations available')

        # Do not remove the conversion to list - fuzzy matching does act up
        # otherwise
        nodes = list(G.nodes)
        if source not in nodes:
            best_match = fw.process.extractOne(source, nodes,
                                               scorer=fw.fuzz.token_sort_ratio)
            raise ValueError(f'Source "{source}" has no known bridging '
                             f'registrations. Did you mean "{best_match[0]}" '
                             'instead?')
        if target not in G.nodes:
            best_match = fw.process.extractOne(target, nodes,
                                               scorer=fw.fuzz.token_sort_ratio)
            raise ValueError(f'Target "{target}" has no known bridging '
                             f'registrations. Did you mean "{best_match[0]}" '
                             'instead?')

        if via:
            via = list(utils.make_iterable(via))  # do not remove the list() here
            for v in via:
                if v not in G.nodes:
                    best_match = fw.process.extractOne(v, nodes,
                                                       scorer=fw.fuzz.token_sort_ratio)
                    raise ValueError(f'Via "{v}" has no known bridging '
                                    f'registrations. Did you mean "{best_match[0]}" '
                                    'instead?')

        if avoid:
            avoid = list(utils.make_iterable(avoid))

        # This will raise a error message if no path is found
        if not via and not avoid:
            try:
                path = nx.shortest_path(G, source, target, weight='weight')
            except nx.NetworkXNoPath:
                raise nx.NetworkXNoPath(f'No bridging path connecting {source} '
                                        f'and {target} found.')
        else:
            # Go through all possible paths and find one that...
            found_any = False  # track if we found any path
            found_good = False  # track if we found a path matching the criteria
            for path in nx.all_simple_paths(G, source, target):
                found_any = True
                # ... has all `via`s...
                if via and all([v in path for v in via]):
                    # ... and none of the `avoid`
                    if avoid:
                        if not any([v in path for v in avoid]):
                            found_good = True
                            break
                    else:
                        found_good = True
                        break
                # If we only have `avoid` but no `via`
                elif avoid and not any([v in path for v in avoid]):
                    found_good = True
                    break

            if not found_any:
                raise nx.NetworkXNoPath(f'No bridging path connecting {source} '
                                        f'and {target} found.')
            elif not found_good:
                if via and avoid:
                    raise nx.NetworkXNoPath(f'No bridging path connecting {source}'
                                            f'and {target} via "{via}" and '
                                            f'avoiding "{avoid}" found')
                elif via:
                    raise nx.NetworkXNoPath(f'No bridging path connecting {source}'
                                            f'and {target} via "{via}" found.')
                else:
                    raise nx.NetworkXNoPath(f'No bridging path connecting {source}'
                                            f'and {target} avoiding "{avoid}" found.')

        # `path` holds the sequence of nodes we are traversing but not which
        # transforms (i.e. edges) to use
        transforms = []
        for n1, n2 in zip(path[:-1], path[1:]):
            this_edges = []
            i = 0
            # First collect all edges between those two nodes
            # - this is annoyingly complicated with MultiDiGraphs
            while True:
                try:
                    e = G.edges[(n1, n2, i)]
                except KeyError:
                    break
                this_edges.append([e['transform'], e['weight']])
                i += 1

            # Now find the edge with the highest weight
            # (inverse transforms might have a lower weight)
            this_edges = sorted(this_edges, key=lambda x: x[-1])
            transforms.append(this_edges[-1][0])

        return path, transforms

    def find_all_bridging_paths(self, source: str,
                                target: str,
                                via: Optional[str] = None,
                                avoid: Optional[str] = None,
                                reciprocal: bool = True,
                                cutoff: int = None) -> tuple:
        """Find all bridging paths from source to target.

        Parameters
        ----------
        source :        str
                        Source from which to transform to `target`.
        target :        str
                        Target to which to transform to.
        via :           str | list thereof, optional
                        Force specific intermediate template(s).
        avoid :         str | list thereof, optional
                        Avoid specific intermediate template(s).
        reciprocal :    bool | float
                        If True or float, will add forward and inverse edges for
                        transforms that are invertible. If float, the inverse
                        edges' weights will be scaled by that factor.
        cutoff :        int, optional
                        Depth to stop the search. Only paths of length
                        <= cutoff are returned.

        Returns
        -------

        path :          list
                        Path from source to target: [source, ..., target]
        transforms :    list
                        Transforms as [[path_to_transform, inverse], ...]

        """
        # Generate (or get cached) bridging graph
        G = self.bridging_graph(reciprocal=reciprocal)

        if len(G) == 0:
            raise ValueError('No bridging registrations available')

        # Do not remove the conversion to list - fuzzy matching does act up
        # otherwise
        nodes = list(G.nodes)
        if source not in nodes:
            best_match = fw.process.extractOne(source, nodes,
                                               scorer=fw.fuzz.token_sort_ratio)
            raise ValueError(f'Source "{source}" has no known bridging '
                             f'registrations. Did you mean "{best_match[0]}" '
                             'instead?')
        if target not in G.nodes:
            best_match = fw.process.extractOne(target, nodes,
                                               scorer=fw.fuzz.token_sort_ratio)
            raise ValueError(f'Target "{target}" has no known bridging '
                             f'registrations. Did you mean "{best_match[0]}" '
                             'instead?')

        if via and via not in G.nodes:
            best_match = fw.process.extractOne(via, nodes,
                                               scorer=fw.fuzz.token_sort_ratio)
            raise ValueError(f'Via "{via}" has no known bridging '
                             f'registrations. Did you mean "{best_match[0]}" '
                             'instead?')

        # This will raise a error message if no path is found
        for path in nx.all_simple_paths(G, source, target, cutoff=cutoff):
            # Skip paths that don't contain `via`
            if isinstance(via, str) and (via not in path):
                continue
            elif isinstance(via, (list, tuple, np.ndarray)) and not all([v in path for v in via]):
                continue

            # Skip paths that contain `avoid`
            if isinstance(avoid, str) and (avoid in path):
                continue
            elif isinstance(avoid, (list, tuple, np.ndarray)) and any([v in path for v in avoid]):
                continue

            # `path` holds the sequence of nodes we are traversing but not which
            # transforms (i.e. edges) to use
            transforms = []
            for n1, n2 in zip(path[:-1], path[1:]):
                this_edges = []
                i = 0
                # First collect all edges between those two nodes
                # - this is annoyingly complicated with MultiDiGraphs
                while True:
                    try:
                        e = G.edges[(n1, n2, i)]
                    except KeyError:
                        break
                    this_edges.append([e['transform'], e['weight']])
                    i += 1

                # Now find the edge with the highest weight
                # (inverse transforms might have a lower weight)
                this_edges = sorted(this_edges, key=lambda x: x[-1])
                transforms.append(this_edges[-1][0])

            yield path, transforms

    @functools.lru_cache()
    def shortest_bridging_seq(self, source: str, target: str,
                              via: Optional[str] = None,
                              inverse_weight: float = .5) -> tuple:
        """Find shortest bridging sequence to get from source to target.

        Parameters
        ----------
        source :            str
                            Source from which to transform to `target`.
        target :            str
                            Target to which to transform to.
        via :               str | list of str
                            Waystations to traverse on the way from source to
                            target.
        inverse_weight :    float
                            Weight for inverse transforms. If < 1 will prefer
                            forward transforms.

        Returns
        -------
        sequence :          (N, ) array
                            Sequence of registrations that will be traversed.
        transform_seq :     TransformSequence
                            Class that collates the required transforms to get
                            from source to target.

        """
        # Generate sequence of nodes we need to find a path for
        # Minimally it's just from source to target
        nodes = np.array([source, target])

        if via:
            nodes = np.insert(nodes, 1, via)

        seq = [nodes[0]]
        transforms = []
        for n1, n2 in zip(nodes[:-1], nodes[1:]):
            path, tr = self.find_bridging_path(n1, n2, reciprocal=inverse_weight)
            seq = np.append(seq, path[1:])
            transforms = np.append(transforms, tr)

        if any(np.unique(seq, return_counts=True)[1] > 1):
            logger.warning('Bridging sequence contains loop: '
                           f'{"->".join(seq)}')

        # Generate the transform sequence
        transform_seq = TransformSequence(*transforms)

        return seq, transform_seq

    def find_mirror_reg(self, template: str, non_found: str = 'raise') -> tuple:
        """Search for a mirror transformation for given template.

        Typically a mirror transformation specifies a non-rigid transformation
        to correct asymmetries in an image.

        Parameters
        ----------
        template :  str
                    Name of the template to find a mirror transformation for.
        non_found : "raise" | "ignore"
                    What to do if no mirror transformation is found. If "ignore"
                    and no mirror transformation found, will silently return
                    `None`.

        Returns
        -------
        tuple
                    Named tuple containing a mirror transformation. Will only
                    ever return one - even if multiple are available.

        """
        for tr in self.mirrors:
            if tr.source == template:
                return tr

        if non_found == 'raise':
            raise ValueError(f'No mirror transformation found for {template}')
        return None

    def find_closest_mirror_reg(self, template: str, non_found: str = 'raise') -> str:
        """Search for the closest mirror transformation for given template.

        Typically a mirror transformation specifies a non-rigid transformation
        to correct asymmetries in an image.

        Parameters
        ----------
        template :  str
                    Name of the template to find a mirror transformation for.
        non_found : "raise" | "ignore"
                    What to do if there is no path to a mirror transformation.
                    If "ignore" and no path is found, will silently return
                    `None`.

        Returns
        -------
        str
                    Name of the closest template with a mirror transform.

        """
        # Templates with mirror registrations
        temps_w_mirrors = [t.source for t in self.mirrors]

        # Add symmetrical template brains
        temps_w_mirrors += [t.label for t in self.templates if getattr(t, 'symmetrical', False) == True]

        if not temps_w_mirrors:
            raise ValueError('No mirror transformations registered')

        # If this template has a mirror registration:
        if template in temps_w_mirrors:
            return template

        # Get bridging graph
        G = self.bridging_graph()

        if template not in G.nodes:
            raise ValueError(f'"{template}" does not appear to be a registered '
                             'template')

        # Get path lengths from template to all other nodes
        pl = nx.single_source_dijkstra_path_length(G, template)

        # Subset to targets that have a mirror reg
        pl = {k: v for k, v in pl.items() if k in temps_w_mirrors}

        # Find the closest mirror
        cl = sorted(pl.keys(), key=lambda x: pl[x])

        # If any, return the closests
        if cl:
            return cl[0]

        if non_found == 'raise':
            raise ValueError(f'No path to a mirror transformation found for "{template}"')

        return None

    def find_template(self, name: str, non_found: str = 'raise') -> 'TemplateBrain':
        """Search for a given template (brain).

        Parameters
        ----------
        name :      str
                    Name of the template to find a mirror transformation for.
                    Searches against `name` and `label` (short name) properties
                    of registered templates.
        non_found : "raise" | "ignore"
                    What to do if no mirror transformation is found. If "ignore"
                    and no mirror transformation found, will silently return
                    `None`.

        Returns
        -------
        TemplateBrain

        """
        for tmp in self.templates:
            if getattr(tmp, 'label', None) == name:
                return tmp
            if getattr(tmp, 'name', None) == name:
                return tmp

        if non_found == 'raise':
            raise ValueError(f'No template brain registered that matches "{name}"')
        return None

    def plot_bridging_graph(self, **kwargs):
        """Draw bridging graph using networkX.

        Parameters
        ----------
        **kwargs
                    Keyword arguments are passed to `networkx.draw_networkx`.

        Returns
        -------
        None

        """
        # Get graph
        G = self.bridging_graph(reciprocal=False)

        # Draw nodes and edges
        node_labels = {n: n for n in G.nodes}
        pos = nx.kamada_kawai_layout(G)

        # Draw all nodes
        nx.draw_networkx_nodes(G, pos=pos, node_color='lightgrey',
                               node_shape='o', node_size=300)
        nx.draw_networkx_labels(G, pos=pos, labels=node_labels,
                                font_color='k', font_size=10)

        # Draw edges by type of transform
        edge_types = set([e[2]['type'] for e in G.edges(data=True)])

        lines = []
        labels = []
        for t, c in zip(edge_types,
                        sns.color_palette('muted', len(edge_types))):
            subset = [e for e in G.edges(data=True) if e[2]['type'] == t]
            nx.draw_networkx_edges(G, pos=pos, edgelist=subset,
                                   edge_color=mcl.to_hex(c), width=1.5)
            lines.append(Line2D([0], [0], color=c, linewidth=2, linestyle='-'))
            labels.append(t)

        plt.legend(lines, labels)


def xform_brain(x: Union['core.NeuronObject', 'pd.DataFrame', 'np.ndarray'],
                source: str,
                target: str,
                via: Optional[str] = None,
                avoid: Optional[str] = None,
                affine_fallback: bool = True,
                caching: bool = True,
                verbose: bool = True) -> Union['core.NeuronObject',
                                               'pd.DataFrame',
                                               'np.ndarray']:
    """Transform 3D data between template brains.

    This requires the appropriate transforms to be registered with `navis`.
    See the docs/tutorials for details.

    Notes
    -----
    For Neurons only: transforms can introduce a change in the units (e.g. if
    the transform goes from micron to nanometer space). Some template brains have
    their units hard-coded in their meta data (as `_navis_units`). If that's
    not the case we fall-back to trying to infer any change in units by comparing
    distances between x/y/z coordinate before and after the transform. That
    approach works reasonably well with base 10 increments (e.g. nm -> um) but
    may be off with odd changes in units (e.g. physical -> voxel space).
    Regardless of whether hard-coded or inferred, any change in units is used to
    update the `.units` property and node/soma radii for TreeNeurons.

    Parameters
    ----------
    x :                 Neuron/List | numpy.ndarray | pandas.DataFrame
                        Data to transform. Dataframe must contain `['x', 'y', 'z']`
                        columns. Numpy array must be shape `(N, 3)`.
    source :            str
                        Source template brain that the data currently is in.
    target :            str
                        Target template brain that the data should be
                        transformed into.
    via :               str | list thereof, optional
                        Optionally set intermediate template(s). This can be
                        helpful to force a specific transformation sequence.
    avoid :             str | list thereof, optional
                        Prohibit going through specific intermediate template(s).
    affine_fallback :   bool
                        In some cases the non-rigid transformation of points
                        can fail - for example if points are outside the
                        deformation field. If that happens, they will be
                        returned as `NaN`. If `affine_fallback=True`
                        we will apply only the rigid affine part of the
                        transformation to those points to get as close as
                        possible to the correct coordinates.
    caching :           bool
                        If True, will (pre-)cache data for transforms whenever
                        possible. Depending on the data and the type of
                        transforms this can speed things up significantly at the
                        cost of increased memory usage:
                          - `False` = no upfront cost, lower memory footprint
                          - `True` = higher upfront cost, most definitely faster
                        Only applies if input is NeuronList and if transforms
                        include H5 transform.
    verbose :           bool
                        If True, will print some useful info on transform.

    Returns
    -------
    same type as `x`
                        Copy of input with transformed coordinates.

    Examples
    --------
    This example requires the
    [flybrains](https://github.com/navis-org/navis-flybrains)
    library to be installed: `pip3 install flybrains`

    Also, if you haven't already, you will need to have the optional Saalfeld
    lab (Janelia Research Campus) transforms installed (this is a one-off):

    >>> import flybrains                                        # doctest: +SKIP
    >>> flybrains.download_jrc_transforms()                     # doctest: +SKIP

    Once `flybrains` is installed and you have downloaded the registrations,
    you can run this:

    >>> import navis
    >>> import flybrains
    >>> # navis example neurons are in raw (8nm voxel) hemibrain (JRCFIB2018Fraw) space
    >>> n = navis.example_neurons(1)
    >>> # Transform to FAFB14 space
    >>> xf = navis.xform_brain(n, source='JRCFIB2018Fraw', target='FAFB14') # doctest: +SKIP

    See Also
    --------
    [`navis.xform`][]
                    Lower level entry point that takes data and applies a given
                    transform or sequence thereof.
    [`navis.mirror_brain`][]
                    Uses non-rigid transforms to mirror neurons from the left
                    to the right side of given template brain and vice versa.

    """
    if not isinstance(source, str):
        TypeError(f'Expected source of type str, got "{type(source)}"')

    if not isinstance(target, str):
        TypeError(f'Expected target of type str, got "{type(target)}"')

    # Get the transformation sequence
    path, transforms = registry.find_bridging_path(source, target, via=via, avoid=avoid)

    if verbose:
        path_str = path[0]
        for p, tr in zip(path[1:], transforms):
            if isinstance(tr, AliasTransform):
                link = '='
            else:
                link = '->'
            path_str += f' {link} {p}'

        print('Transform path:', path_str)

    # Combine into transform sequence
    trans_seq = TransformSequence(*transforms)

    # Apply transform and returned xformed points
    xf = xform(x, transform=trans_seq, caching=caching,
               affine_fallback=affine_fallback)

    # We might be able to set the correct units based on the target template's
    # meta data (the "guessed" new units can be off if the transform is
    # not base 10 which happens for e.g. voxels -> physical space)
    if isinstance(xf, (core.NeuronList, core.BaseNeuron)):
        # First we need to find the last non-alias template space
        for tmp, tr in zip(path[::-1], transforms[::-1]):
            if not isinstance(tr, AliasTransform):
                # There is a chance that there is no meta data for this template
                try:
                    last_temp = registry.find_template(tmp)
                except ValueError:
                    break
                except BaseException:
                    raise
                # If this template brain has a property for navis units
                if hasattr(last_temp, '_navis_units'):
                    for n in core.NeuronList(xf):
                        n.units = last_temp._navis_units
                break

    return xf


def _guess_change(xyz_before: np.ndarray,
                  xyz_after: np.ndarray,
                  sample: float = .1) -> tuple:
    """Guess change in units during xforming."""
    if isinstance(xyz_before, pd.DataFrame):
        xyz_before = xyz_before[['x', 'y', 'z']].values
    if isinstance(xyz_after, pd.DataFrame):
        xyz_after = xyz_after[['x', 'y', 'z']].values

    # Select the same random sample of points in both spaces
    if sample <= 1:
        sample = int(xyz_before.shape[0] * sample)
    rnd_ix = np.random.choice(xyz_before.shape[0], sample, replace=False)
    sample_bef = xyz_before[rnd_ix, :]
    sample_aft = xyz_after[rnd_ix, :]

    # Get pairwise distance between those points
    dist_pre = pdist(sample_bef)
    dist_post = pdist(sample_aft)

    # Calculate how the distance between nodes changed and get the average
    # Note we are ignoring nans - happens e.g. when points did not transform.
    with np.errstate(divide='ignore', invalid='ignore'):
        change = dist_post / dist_pre
    # Drop infinite values in rare cases where nodes end up on top of another
    mean_change = np.nanmean(change[change < np.inf])

    # Find the order of magnitude
    magnitude = round(math.log10(mean_change))

    return mean_change, magnitude


def symmetrize_brain(x: Union['core.NeuronObject', 'pd.DataFrame', 'np.ndarray'],
                     template: Union[str, 'TemplateBrain'],
                     via: Optional[str] = 'auto',
                     verbose: bool = False) -> Union['core.NeuronObject',
                                                     'pd.DataFrame',
                                                     'np.ndarray']:
    """Symmetrize 3D object (neuron, coordinates).

    The way this works is by:
     1. Finding the closest mirror transform (unless provided)
     2. Mirror data on the left-hand-side to the right-hand-side using the
        proper (warp) mirror transform to offset deformations
     3. Simply flip that data back to the left-hand-side

    This works reasonably well but may produce odd results around the midline.
    For high quality symmetrization you are better off generating dedicated
    transform (see `navis-flybrains` for an example).

    Parameters
    ----------
    x :             Neuron/List | Volume/trimesh | numpy.ndarray | pandas.DataFrame
                    Data to transform. Dataframe must contain `['x', 'y', 'z']`
                    columns. Numpy array must be shape `(N, 3)`.
    template :      str | TemplateBrain
                    Source template brain space that the data is in. If string
                    will be searched against registered template brains.
    via :           "auto" | str
                    By default ("auto") it will find and apply the closest
                    mirror transform. You can also specify a template that
                    should be used. That template must have a mirror transform!
    verbose :       bool
                    If True, will print some useful info on the transform(s).

    Returns
    -------
    xs
                    Same object type as input (array, neurons, etc) but
                    hopefully symmetrical.

    Examples
    --------
    This example requires the
    [flybrains](https://github.com/navis-org/navis-flybrains)
    library to be installed: `pip3 install flybrains`

    >>> import navis
    >>> import flybrains
    >>> # Get the FAFB14 neuropil mesh
    >>> m = flybrains.FAFB14.mesh
    >>> # Symmetrize the mesh
    >>> s = navis.symmetrize_brain(m, template='FAFB14')
    >>> # Plot side-by-side for comparison
    >>> m.plot3d()                                              # doctest: +SKIP
    >>> s.plot3d(color=(1, 0, 0))                               # doctest: +SKIP

    """
    if not isinstance(template, str):
        TypeError(f'Expected template of type str, got "{type(template)}"')

    if via == 'auto':
        # Find closest mirror transform
        via = registry.find_closest_mirror_reg(template)

    if isinstance(x, core.NeuronList):
        if len(x) == 1:
            x = x[0]
        else:
            xf = []
            for n in config.tqdm(x, desc='Mirroring',
                                 disable=config.pbar_hide,
                                 leave=config.pbar_leave):
                xf.append(symmetrize_brain(n,
                                           template=template,
                                           via=via))
            return core.NeuronList(xf)

    if isinstance(x, core.BaseNeuron):
        x = x.copy()
        if isinstance(x, core.TreeNeuron):
            x.nodes = symmetrize_brain(x.nodes,
                                       template=template,
                                       via=via)
        elif isinstance(x, core.Dotprops):
            x.points = symmetrize_brain(x.points,
                                        template=template,
                                        via=via)
            # Set tangent vectors and alpha to None so they will be regenerated
            x._vect = x._alpha = None
        elif isinstance(x, core.MeshNeuron):
            x.vertices = symmetrize_brain(x.vertices,
                                          template=template,
                                          via=via)
        else:
            raise TypeError(f"Don't know how to transform neuron of type '{type(x)}'")

        if x.has_connectors:
            x.connectors = symmetrize_brain(x.connectors,
                                            template=template,
                                            via=via)
        return x
    elif isinstance(x, tm.Trimesh):
        x = x.copy()
        x.vertices = symmetrize_brain(x.vertices,
                                      template=template,
                                      via=via)
        return x
    elif isinstance(x, pd.DataFrame):
        if any([c not in x.columns for c in ['x', 'y', 'z']]):
            raise ValueError('DataFrame must have x, y and z columns.')
        x = x.copy()
        x[['x', 'y', 'z']] = symmetrize_brain(
            x[['x', 'y', 'z']].values.astype(float),
            template=template,
            via=via
        )
        return x
    else:
        try:
            # At this point we expect numpy arrays
            x = np.asarray(x)
        except BaseException:
            raise TypeError(f'Unable to transform data of type "{type(x)}"')

        if not x.ndim == 2 or x.shape[1] != 3:
            raise ValueError('Array must be of shape (N, 3).')

    # Now find the meta info for this template brain
    if isinstance(template, TemplateBrain):
        tb = template
    else:
        tb = registry.find_template(template, non_found='raise')

    # Get the bounding box
    if not hasattr(tb, 'boundingbox'):
        raise ValueError(f'Template "{tb.label}" has no bounding box info.')

    if not isinstance(tb.boundingbox, (list, tuple, np.ndarray)):
        raise TypeError("Expected the template brain's bounding box to be a "
                        f"list, tuple or array - got '{type(tb.boundingbox)}'")

    # Get bounding box of template brain
    bbox = np.asarray(tb.boundingbox)

    # Reshape if flat array
    if bbox.ndim == 1:
        bbox = bbox.reshape(3, 2)

    # Find points on the left
    center = bbox[0][0] + (bbox[0][1] - bbox[0][0]) / 2
    is_left = x[:, 0] > center

    # Make a copy of the original data
    x = x.copy()

    # If nothing to symmetrize - return
    if is_left.sum() == 0:
        return x

    # Mirror with compensation for deformations
    xm = mirror_brain(x[is_left], template=template, via=via,
                      mirror_axis='x', verbose=verbose)

    # And now flip them back without compensation for deformations
    xmf = mirror_brain(xm, template=template, warp=False, mirror_axis='x')

    # Replace values
    x[is_left] = xmf

    return x


def mirror_brain(x: Union['core.NeuronObject', 'pd.DataFrame', 'np.ndarray'],
                 template: Union[str, 'TemplateBrain'],
                 mirror_axis: Union[Literal['x'],
                                    Literal['y'],
                                    Literal['z']] = 'x',
                 warp: Union[Literal['auto'], bool] = 'auto',
                 via: Optional[str] = None,
                 verbose: bool = False) -> Union['core.NeuronObject',
                                                 'pd.DataFrame',
                                                 'np.ndarray']:
    """Mirror 3D object (neuron, coordinates) about given axis.

    The way this works is:
     1. Look up the length of the template space along the given axis. For this,
        the template space has to be registered (see docs for details).
     2. Flip object along midpoint of axis using a affine transformation.
     3. (Optional) Apply a warp transform that corrects asymmetries.

    Parameters
    ----------
    x :             Neuron/List | Volume/trimesh | numpy.ndarray | pandas.DataFrame
                    Data to transform. Dataframe must contain `['x', 'y', 'z']`
                    columns. Numpy array must be shape `(N, 3)`.
    template :      str | TemplateBrain
                    Source template brain space that the data is in. If string
                    will be searched against registered template brains.
                    Alternatively check out [`navis.transforms.mirror`][]
                    for a lower level interface.
    mirror_axis :   'x' | 'y' | 'z', optional
                    Axis to mirror. Defaults to `x`.
    warp :          bool | "auto" | Transform, optional
                    If 'auto', will check if a non-rigid mirror transformation
                    exists for the given `template` and apply it after the
                    flipping. Alternatively, you can also pass a Transform or
                    TransformSequence directly.
    via :           str | None
                    If provided, (e.g. "FCWB") will first transform coordinates
                    into that space, then mirror and transform back.
                    Use this if there is no mirror registration for the original
                    template, or to transform to a symmetrical template in which
                    flipping is sufficient.
    verbose :       bool
                    If True, will print some useful info on the transform(s).

    Returns
    -------
    xf
                    Same object type as input (array, neurons, etc) but with
                    transformed coordinates.

    Examples
    --------
    This example requires the
    [flybrains](https://github.com/navis-org/navis-flybrains)
    library to be installed: `pip3 install flybrains`

    Also, if you haven't already, you will need to have the optional Saalfeld
    lab (Janelia Research Campus) transforms installed (this is a one-off):

    >>> import flybrains                                        # doctest: +SKIP
    >>> flybrains.download_jrc_transforms()                     # doctest: +SKIP

    Once `flybrains` is installed and you have downloaded the registrations,
    you can run this:

    >>> import navis
    >>> import flybrains
    >>> # navis example neurons are in raw hemibrain (JRCFIB2018Fraw) space
    >>> n = navis.example_neurons(1)
    >>> # Mirror about x axis (this is a simple flip in this case)
    >>> mirrored = navis.mirror_brain(n * 8 / 1000, tem plate='JRCFIB2018F', via='JRC2018F') # doctest: +SKIP
    >>> # We also need to get back to raw coordinates
    >>> mirrored = mirrored / 8 * 1000                          # doctest: +SKIP

    See Also
    --------
    [`navis.mirror`][]
                    Lower level function for mirroring. You can use this if
                    you want to mirror data without having a registered
                    template for it.

    """
    utils.eval_param(mirror_axis, name='mirror_axis',
                     allowed_values=('x', 'y', 'z'), on_error='raise')
    if not isinstance(warp, (BaseTransform, TransformSequence)):
        utils.eval_param(warp, name='warp',
                         allowed_values=('auto', True, False), on_error='raise')

    # If we go via another brain space
    if via and via != template:
        # Xform to "via" space
        xf = xform_brain(x, source=template, target=via, verbose=verbose)
        # Mirror
        xfm = mirror_brain(xf,
                           template=via,
                           mirror_axis=mirror_axis,
                           warp=warp,
                           via=None)
        # Xform back to original template space
        xfm_inv = xform_brain(xfm, source=via, target=template, verbose=verbose)
        return xfm_inv

    if isinstance(x, core.NeuronList):
        if len(x) == 1:
            x = x[0]
        else:
            xf = []
            for n in config.tqdm(x, desc='Mirroring',
                                 disable=config.pbar_hide,
                                 leave=config.pbar_leave):
                xf.append(mirror_brain(n,
                                       template=template,
                                       mirror_axis=mirror_axis,
                                       warp=warp))
            return core.NeuronList(xf)

    if isinstance(x, core.BaseNeuron):
        x = x.copy()
        if isinstance(x, core.TreeNeuron):
            x.nodes = mirror_brain(x.nodes,
                                   template=template,
                                   mirror_axis=mirror_axis,
                                   warp=warp)
        elif isinstance(x, core.Dotprops):
            if isinstance(x.k, type(None)) or x.k <= 0:
                # If no k, we need to mirror vectors too. Note that this is less
                # than ideal though! Here, we are scaling the vector by the
                # dotprop's sampling resolution (i.e. ideally a representative
                # distance between the points) because if the vectors are too
                # small any warping transform will make them go haywire
                hp = mirror_brain(x.points + x.vect * x.sampling_resolution * 2,
                                  template=template,
                                  mirror_axis=mirror_axis,
                                  warp=warp)

            x.points = mirror_brain(x.points,
                                    template=template,
                                    mirror_axis=mirror_axis,
                                    warp=warp)

            if isinstance(x.k, type(None)) or x.k <= 0:
                # Re-generate vectors
                vect = x.points - hp
                vect = vect / np.linalg.norm(vect, axis=1).reshape(-1, 1)
                x._vect = vect
            else:
                # Set tangent vectors and alpha to None so they will be
                # regenerated on demand
                x._vect = x._alpha = None
        elif isinstance(x, core.MeshNeuron):
            x.vertices = mirror_brain(x.vertices,
                                      template=template,
                                      mirror_axis=mirror_axis,
                                      warp=warp)
            # We also need to flip the normals
            x.faces = x.faces[:, ::-1]
        else:
            raise TypeError(f"Don't know how to transform neuron of type '{type(x)}'")

        if x.has_connectors:
            x.connectors = mirror_brain(x.connectors,
                                        template=template,
                                        mirror_axis=mirror_axis,
                                        warp=warp)
        return x
    elif isinstance(x, tm.Trimesh):
        x = x.copy()
        x.vertices = mirror_brain(x.vertices,
                                  template=template,
                                  mirror_axis=mirror_axis,
                                  warp=warp)

        # We also need to flip the normals
        x.faces = x.faces[:, ::-1]
        return x
    elif isinstance(x, pd.DataFrame):
        if any([c not in x.columns for c in ['x', 'y', 'z']]):
            raise ValueError('DataFrame must have x, y and z columns.')
        x = x.copy()
        x[["x", "y", "z"]] = mirror_brain(
            x[["x", "y", "z"]].values,
            template=template,
            mirror_axis=mirror_axis,
            warp=warp,
        )
        return x
    else:
        try:
            # At this point we expect numpy arrays
            x = np.asarray(x)
        except BaseException:
            raise TypeError(f'Unable to transform data of type "{type(x)}"')

        if not x.ndim == 2 or x.shape[1] != 3:
            raise ValueError('Array must be of shape (N, 3).')

    if not isinstance(template, str):
        TypeError(f'Expected template of type str, got "{type(template)}"')

    if isinstance(warp, (BaseTransform, TransformSequence)):
        mirror_trans = warp
    elif warp:
        # See if there is a mirror registration
        mirror_trans = registry.find_mirror_reg(template, non_found='ignore')

        # Get actual transform from tuple
        if mirror_trans:
            mirror_trans = mirror_trans.transform
        # If warp was not "auto" and we didn't find a registration, raise
        elif warp != 'auto' and not mirror_trans:
            raise ValueError(f'No mirror transform found for "{template}"')
    else:
        mirror_trans = None

    # Now find the meta info about the template brain
    if isinstance(template, TemplateBrain):
        tb = template
    else:
        tb = registry.find_template(template, non_found='raise')

    # Get the bounding box
    if not hasattr(tb, 'boundingbox'):
        raise ValueError(f'Template "{tb.label}" has no bounding box info.')

    if not isinstance(tb.boundingbox, (list, tuple, np.ndarray)):
        raise TypeError("Expected the template brain's bounding box to be a "
                        f"list, tuple or array - got '{type(tb.boundingbox)}'")

    # Get bounding box of template brain
    bbox = np.asarray(tb.boundingbox)

    # Reshape if flat array
    if bbox.ndim == 1:
        bbox = bbox.reshape(3, 2)

    # Index of mirror axis
    ix = {'x': 0, 'y': 1, 'z': 2}[mirror_axis]

    if bbox.shape == (3, 2):
        # In nat.templatebrains this is using the sum (min+max) but have a
        # suspicion that this should be the difference (max-min)
        mirror_axis_size = bbox[ix, :].sum()
    elif bbox.shape == (2, 3):
        mirror_axis_size = bbox[:, ix].sum()
    else:
        raise ValueError('Expected bounding box to be of shape (3, 2) or (2, 3)'
                         f' got {bbox.shape}')

    return mirror(x, mirror_axis=mirror_axis, mirror_axis_size=mirror_axis_size,
                  warp=mirror_trans)


class TemplateBrain:
    """Generic base class for template brains.

    Minimally, a template should have a `name` and `label` property. For
    mirroring, it also needs a `boundingbox`.

    See [flybrains](https://github.com/navis-org/navis-flybrains) for
    an example of how to use template brains.

    """
    def __init__(self, **properties):
        """Initialize class."""
        for k, v in properties.items():
            setattr(self, k, v)

    @property
    def mesh(self):
        """Mesh represenation of this brain."""
        if not hasattr(self, '_mesh'):
            name = getattr(self, 'regName', getattr(self, 'name', None))
            raise ValueError(f'{name} does not appear to have a mesh')
        return self._mesh


# Initialize the registry
registry = TemplateRegistry()



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.

import warnings

import numpy as np

from inspect import signature

from .. import core, utils, config

from .moving_least_squares import MovingLeastSquaresTransform

logger = config.logger


def align_pairwise(x, y=None, method='rigid', sample=None, progress=True, **kwargs):
    """Run a pairwise alignment between given neurons.

    Requires the `pycpd` library.

    Parameters
    ----------
    x :         navis.NeuronList
                Neurons to align to other neurons.
    y :         navis.NeuronList, optional
                The neurons to align to. If `None`, will run pairwise
                alignment of `x` vs `x`.
    method :    "rigid" | "deform" | "pca" | "rigid+deform"
                Which method to use for alignment. Maps to the respective
                `navis.align_{method}` function. "rigid+deform" performs a
                rigid followed by a warping alignment.
    sample :    float [0-1], optional
                If provided, will calculate an initial registration on only
                the given fraction of points followed by a landmark transform
                to transform the rest. Use this to speed things up.
    **kwargs
                Keyword arguments are passed through to the respective
                alignment function.

    Returns
    -------
    np.ndarray
                Array of shape (x, y) with the pairwise-aligned neurons.

    See Also
    --------
    [`navis.nblast_align`][]
                Runs an NBLAST where neurons are first aligned pairwise.

    Examples
    --------
    >>> import navis
    >>> nl = navis.example_neurons(2, kind='skeleton')
    >>> aligned = navis.align.align_pairwise(nl, method='rigid', sample=.2)

    """
    if y is None:
        y = x

    utils.eval_param(x, name='x', allowed_types=(core.NeuronList, ))
    utils.eval_param(y, name='y', allowed_types=(core.NeuronList, ))
    utils.eval_param(method, name='method',
                     allowed_values=('rigid', 'deform', 'pca'))

    func = {'rigid': align_rigid,
            'deform': align_deform,
            'pca': align_pca,
            'rigid+deform': _align_rigid_deform}[method]

    aligned = []
    for n1 in config.tqdm(x,
                          desc='Aligning',
                          disable=not progress or len(x) == 1):
        aligned.append([])
        for n2 in y:
            if n1 is n2:
                xf = n1
            else:
                xf = func(n1, target=n2, sample=sample, progress=False, **kwargs)[0][0]
            aligned[-1].append(xf)

    return np.array(aligned)


def _align_rigid_deform(x, target, sample=None, progress=True, **kwargs):
    """Thin wrapper to perform a rigid followed by a non-rigid alignment.

    Examples
    --------
    # For doctests
    >>> import navis
    >>> n1, n2 = navis.example_neurons(2, kind='skeleton')
    >>> n1_aligned = navis.transforms.align._align_rigid_deform(n1, n2, scale=True)

    """
    # Parse kwargs
    rigid_kwargs = {k: v for k, v in kwargs.items() if k in signature(align_rigid).parameters}
    deform_kwargs = {k: v for k, v in kwargs.items() if k not in signature(align_rigid).parameters}

    xf, _ = align_rigid(x, target, sample=sample, progress=progress, **rigid_kwargs)
    xf2, _ = align_deform(xf, target, sample=sample, progress=progress, **deform_kwargs)

    return xf2


def align_rigid(x, target=None, scale=False, w=0, verbose=False, sample=None, progress=True):
    """Align neurons using a rigid registration.

    Requires the `pycpd` library.

    Parameters
    ----------
    x :             navis.NeuronList
                    Neurons to align.
    target :        navis.Neuron | np.ndarray
                    The neuron that all neurons in `x` will be aligned to.
                    If `None`, neurons will be aligned to the first neuron in `x`!
    scale :         bool
                    If True, will also scale the neuron.
    w :             float
                    `w` is used to account for outliers: higher w = more forgiving.
                    The default is w=0 which can lead to failure to converge on a
                    solution (in particular when scale=False). In that case we
                    incrementally increase `w` by a factor of 10 until we find
                    a solution. Set `verbose=True` to get detailed feedback
                    on the solution.
    sample :        float [0-1], optional
                    If provided, will calculate an initial registration on only
                    the given fraction of points followed by a landmark transform
                    to transform the rest. Use this to speed things up.
    progress :      bool
                    Whether to show a progress bar.

    Returns
    -------
    xf :    navis.NeuronList
            The aligned neurons.
    regs :  list
            The pycpd registration objects.

    Examples
    --------
    >>> import navis
    >>> n1, n2 = navis.example_neurons(2, kind='skeleton')
    >>> n1_aligned, regs = navis.align.align_rigid(n1, n2, sample=.2)

    """
    try:
        from pycpd import RigidRegistration as Registration
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            '`align_rigid()` requires the `pycpd` library:\n'
            '  pip3 install git+https://github.com/siavashk/pycpd@master -U'
            )

    if isinstance(x, core.BaseNeuron):
        x = core.NeuronList(x)

    assert isinstance(x, core.NeuronList)

    if target is None:
        target = x[0]

    target_co = _extract_coords(target)

    # This wraps the registration process
    register = _reg_subsample(Registration, sample=sample)

    xf = x.copy()
    regs = []
    for n in config.tqdm(xf,
                         disable=(not progress) or (len(xf) == 1),
                         desc='Aligning'):
        if n is target:
            continue
        # `w` is used to account for outliers -> higher w = more forgiving
        # the default is w=0 which can lead to failure to converge on a solution
        # in particular when scale=False
        # Our work-around here is to start at w=0 and incrementally increase w
        # if we fail to converge
        # Also note that pycpd ignores the `scale` in earlier versions. The
        # version on PyPI is currently outdated. From what I understand we need
        # the Github version.
        converged = False
        while w <= 0.001:
            try:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    TY, params, reg = register(X=target_co,
                                               Y=_extract_coords(n),
                                               scale=scale,
                                               s=1,
                                               w=w)
                _set_coords(n, TY)
                converged = True
                regs.append(reg)
                break
            except np.linalg.LinAlgError:
                if w == 0:
                    w += 0.000000001
                else:
                    w *= 10

        if verbose:
            if not converged:
                logger.info(f'Registration of {n.id} onto {target.id} did not converge')
            else:
                logger.info(f'Registration of {n.id} onto {target.id} converged for w={w}')

    return xf, regs


def align_deform(x, target=None, sample=None, progress=True, **kwargs):
    """Align neurons using a deformable registration.

    Requires the `pycpd` library. Note that it's often beneficial to first
    run a rough affine alignment via `rigid_align`. Anecdotally, this works
    well to align backbones but tends to pull denser parts (e.g. dendrites)
    into a tight ball.

    Parameters
    ----------
    x :             navis.NeuronList
                    Neurons to align.
    target :        navis.Neuron | np.ndarray
                    The neuron that all neurons in `x` will be aligned to.
                    If `None`, neurons will be aligned to the first neuron in `x`!
    sample :        float [0-1], optional
                    If provided, will calculate an initial registration on only
                    the given fraction of points followed by a landmark transform
                    to transform the rest. Use this to speed things up.
    progress :      bool
                    Whether to show a progress bar.
    **kwargs
                    Additional keyword-argumens are passed through to
                    pycpd.DeformableRegistration. In brief: lower `alpha` and
                    higher `beta` typically make for more fitting deform. I have
                    gone as far as alpha=.01 and beta=10000.

    Returns
    -------
    xf :    navis.NeuronList
            The aligned neurons.
    regs :  list
            The pycpd registration objects.

    Examples
    --------
    >>> import navis
    >>> n1, n2 = navis.example_neurons(2, kind='skeleton')
    >>> n1_aligned, regs = navis.align.align_deform(n1, n2, sample=.2)

    """
    try:
        from pycpd import DeformableRegistration as Registration
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            '`align_deform()` requires the `pycpd` library:\n'
            '  pip3 install git+https://github.com/siavashk/pycpd@master -U'
            )

    if isinstance(x, core.BaseNeuron):
        x = core.NeuronList(x)

    assert isinstance(x, core.NeuronList), f"Expected NeuronList, got {type(x)}"

    if target is None:
        target = x[0]

    target_co = _extract_coords(target)

    # This wraps the registration process
    register = _reg_subsample(Registration, sample=sample)

    # pycpd's deformable registration is very sensitive to the scale of the
    # data. We will hence normalize the neurons to be within the -1 to 1 range
    scale_factor = 0
    for n in x:
        co = _extract_coords(n)
        mx = np.abs(co).max()
        scale_factor = mx if mx > scale_factor else scale_factor

    xf = x / scale_factor
    target_co = target_co / scale_factor
    regs = []
    for n in config.tqdm(xf,
                         disable=(not progress) or (len(xf) == 1),
                         desc='Aligning'):
        if n is target:
            continue
        TY, params, reg = register(X=target_co, Y=_extract_coords(n), **kwargs)
        _set_coords(n, TY)
        regs.append(reg)

    return xf * scale_factor, regs


def align_pca(x, individually=True):
    """Align neurons along their first principal components.

    This will in effect turn the neurons into a 1-dimensional line.
    Requires the `scikit-learn` library.

    Parameters
    ----------
    x :             navis.NeuronList | np.ndarray
                    The neurons to align.
    individually :  bool
                    Whether to align neurons along their individual or
                    collective first principical component.

    Returns
    -------
    xf :    navis.NeuronList
            The PCA-aligned neurons.
    pcas :  list
            The scikit-learn PCA object(s)

    Examples
    --------
    >>> import navis
    >>> n1, n2 = navis.example_neurons(2, kind='skeleton')
    >>> n1_aligned, pcas = navis.align.align_pca(n1, n2)

    """
    try:
        from sklearn.decomposition import PCA
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            '`align_pca()` requires the `scikit-learn` library:\n'
            '  pip3 install scikit-learn -U'
            )

    if isinstance(x, core.BaseNeuron):
        x = core.NeuronList(x)

    assert isinstance(x, core.NeuronList)

    pcas = []
    if not individually:
        # Collect coordinates
        co = [_extract_coords(n) for n in x]
        n_points = [len(c) for c in co]  # track how many points per neuron
        co = np.vstack(co)

        pca = PCA(n_components=1)
        co_new = pca.fit_transform(X=co)

        xf = x.copy()
        i = 0
        for n, le in zip(xf, n_points):
            _set_coords(n, co_new[i: i + le])
            i += le
        pcas.append(pca)
    else:
        xf = x.copy()
        for n in xf:
            pca = PCA(n_components=1)
            _set_coords(n, pca.fit_transform(X=_extract_coords(n)))
            pcas.append(pca)
    return xf, pcas


def _extract_coords(n):
    """Extract xyz coordinates from given object."""
    if isinstance(n, np.ndarray):
        return n
    elif isinstance(n, core.MeshNeuron):
        return n.vertices
    elif isinstance(n, core.TreeNeuron):
        return n.nodes[['x', 'y', 'z']].values
    elif isinstance(n, core.Dotprops):
        return n.points
    else:
        raise TypeError(f'Unable to extract coordinates from {type(n)}')


def _set_coords(n, new_co):
    """Set new xyz coordinates for given object."""
    if new_co.ndim == 2 and new_co.shape[1] == 1:
        new_co = new_co.flatten()

    if new_co.ndim == 2:
        if isinstance(n, core.MeshNeuron):
            n.vertices = new_co
        elif isinstance(n, core.TreeNeuron):
            n.nodes[['x', 'y', 'z']] = new_co
        elif isinstance(n, core.Dotprops):
            n.points = new_co
        else:
            raise TypeError(f'Unable to extract coordinates from {type(n)}')
    # If this is a single vector
    else:
        if isinstance(n, core.MeshNeuron):
            for i in range(3):
                n.vertices[:, i] = new_co
        elif isinstance(n, core.TreeNeuron):
            for i in 'xyz':
                n.nodes[i] = new_co
        elif isinstance(n, core.Dotprops):
            for i in range(3):
                n.points[:, i] = new_co
        else:
            raise TypeError(f'Unable to extract coordinates from {type(n)}')


def _reg_subsample(Registration, sample):
    """Find and apply a transform for a subset of points. Then use a landmark
    transform to move the rest."""
    if sample is not None:
        assert (sample > 0) and (sample <= 1), '`sample` must be >0 and <1'
    def inner(X, Y, **kwargs):
        if sample is not None and (sample != 1):
            # Subsample points
            XS = X[::int(1 / sample)]
            YS = Y[::int(1 / sample)]

            # Find transform for subset of points
            reg = Registration(X=XS, Y=YS, **kwargs)
            TYS, params = reg.register()

            # Make transform from registered points
            tr = MovingLeastSquaresTransform(YS, TYS)

            # Apply transform to the whole set
            TY = tr.xform(Y)
        else:
            reg = Registration(X=X, Y=Y, **kwargs)
            TY, params = reg.register()

        return TY, params, reg

    return inner


#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.

"""Interface with VirtualFlyBrain.org using the vfb_connect library."""

from textwrap import dedent

try:
    from vfb_connect.cross_server_tools import VfbConnect
    vc = VfbConnect(neo_endpoint='http://pdb.v4.virtualflybrain.org', neo_credentials=('neo4j', 'vfb'))
except ModuleNotFoundError:
    msg = dedent("""
          vfb_connect library not found. Please install using pip:

                pip install vfb_connect

          """)
    raise ModuleNotFoundError(msg)
except BaseException:
    raise

import requests

import numpy as np
import pandas as pd

from concurrent.futures import ThreadPoolExecutor, as_completed
from io import StringIO

from .. import config, utils, core

logger = config.get_logger(__name__)


# For convenience some shorthands for datasets
DATABASES = {'hemibrain:1.0.1': 'neuprint_JRC_Hemibrain_1point0point1',
             'hemibrain:v1.1': 'neuprint_JRC_Hemibrain_1point1',
             'hemibrain:v1.1.1': 'neuprint_JRC_Hemibrain_1point1',  # these are effectively the same
             'hemibrain': 'neuprint_JRC_Hemibrain_1point1',         # these are effectively the same
             'FAFB': 'catmaid_FAFB'}


def get_vfb_ids(x, database=None, attach=True):
    """Fetch VFB ID for given neuron(s).

    Parameters
    ----------
    x :         Neuron/List | int | str | iterable
                Neurons, IDs or a list thereof (see Examples). If neuron(s) will
                assume that their IDs correspond to the `accession` (=external)
                ID in VFB.
    database :  str
                Database `x` belongs to. This can be, for example:
                    - 'FAFB'
                    - 'neuronbridge'
                    - 'hemibrain:v1.0.1'
                    - 'hemibrain:v1.1.0'
                    - 'FlyCircuit'
                Use `vfb.vc.neo_query_wrapper.get_dbs()` to get a full list.

    Returns
    -------
    pandas.DataFrame

    Examples
    --------

    >>> import navis
    >>> from navis.interfaces import vfb
    >>> # Example neurons are from the hemibrain dataset
    >>> nl = navis.example_neurons(2)
    >>> # Let's see if VFB knows their IDs ("body IDs" in neuPrint speak)
    >>> vfb_ids = vfb.get_vfb_ids(nl)
    >>> vfb_ids
                                            1734350788    1734350908
    neuronbridge                          VFB_jrchjtdf  VFB_jrchjtdb
    neuprint_JRC_Hemibrain_1point0point1  VFB_jrch078z  VFB_jrch075v
    neuprint_JRC_Hemibrain_1point1        VFB_jrchjtdf  VFB_jrchjtdb

    """
    database = DATABASES.get(database, database)

    if isinstance(x, (core.NeuronList, core.BaseNeuron)):
        x = x.id

    x = utils.make_iterable(x).astype(str)

    res = vc.neo_query_wrapper.xref_2_vfb_id(acc=x, db=database)

    res = {keys: {v['db']: v['vfb_id'] for v in values} for keys, values in res.items()}

    return pd.DataFrame.from_records(res)


def get_vfb_meta(x, database, raw=False):
    """Fetch VFB's meta data for given neuron(s).

    Parameters
    ----------
    x :         Neuron/List | int | str | iterable
                Neurons, IDs or a list thereof (see Examples). If neuron(s) will
                assume that their IDs correspond to the `accession` (=external)
                ID in VFB.
    database :  str
                Database `x` belongs to. This can be, for example:
                    - 'FAFB'
                    - 'neuronbridge'
                    - 'hemibrain:v1.0.1'
                    - 'hemibrain:v1.1.0'
                    - 'FlyCircuit'
                Use `vfb.vc.neo_query_wrapper.get_dbs()` to get a full list.
    raw :       bool
                If True, will return the raw JSON (list of nested dicts).
                If False (default) will return a filtered, flattened DataFrame
                with the most relevant info.

    Returns
    -------
    meta :      pandas.DataFrame

    Examples
    --------


    """
    vfb_ids = get_vfb_ids(x, database=database)

    if isinstance(x, (core.NeuronList, core.BaseNeuron)):
        ids = x.id
    else:
        ids = x

    ids = utils.make_iterable(ids).astype(str)

    miss = ~np.isin(ids, vfb_ids.columns)
    if any(miss):
        logger.warning(f'Did not find VFB meta data for {ids[miss]}')

    to_fetch = vfb_ids.iloc[0][ids[~miss]].values.astype(str)

    json = vc.neo_query_wrapper.get_TermInfo(to_fetch.tolist())

    if raw:
        return json

    meta = []

    for n in json:
        # Some core info
        label = n['term']['core']['label']
        short = n['term']['core']['short_form']
        desc = n['term']['description']
        comment = n['term']['comment']

        # Xref info
        acc = list(set([ref['accession'] for ref in n['xrefs']]))
        sources = [ref['site']['label'] for ref in n['xrefs']]

        # Get available templates
        temps = [im['image']['template_anatomy']['label'] for im in n['channel_image']]
        temps = list(set(temps))

        # Find neurotransmitter (not sure this covers all the terms)
        for trans in ('Cholinergic', 'GABAergic', 'Glutamatergic',
                      'Serotonergic', 'Octopaminergic', 'NA'):
            if trans in n['term']['core']['types']:
                break

        meta.append([label, short, trans, acc, sources, desc, comment, temps])

    meta = pd.DataFrame(meta, columns=['label', 'vfb_id', 'transmitter',
                                       'accession_id', 'sources', 'description',
                                       'comments', 'templates'])

    return meta


def get_skeletons(x, template=None, max_threads=5, verbose=True):
    """Fetch skeletons for given VFB IDs.

    Parameters
    ----------
    x :             str | list thereof
                    VFB IDs to fetch skeletons for.
    template :      str, optional
                    The template space the skeletons should be in. Use
                    `vfb.vc.neo_query_wrapper.get_templates()` to get a fulll
                    list of available templates. Not specifying a template will
                    return all available images.
    max_threads :   int, optional
                    Max number of parallel requests.

    Returns
    -------
    NeuronList

    Examples
    --------
    Fetch skeletons based on their VFB ID (e.g. copy-pasted from website)

    >>> from navis.interfaces import vfb
    >>> n = vfb.get_skeletons('VFB_jrchjtdf', template='JRC2018Unisex')

    Fetch all available skeletons for the "adult antennal lobe projection neuron DA1 lPN"
    class:

    >>> from navis.interfaces import vfb
    >>> nl = vfb.get_skeletons('FBbt_00067363')

    """
    x = utils.make_iterable(x).astype(str)

    # Grab the meta data
    images = _query_images(x, template=template, verbose=verbose)

    if not images:
        return

    with ThreadPoolExecutor(max_workers=1 if not max_threads else max_threads) as executor:
        futures = {}
        for img in images:
            f = executor.submit(_fetch_single_skeleton, img)
            futures[f] = img['template_channel']['short_form']

        nl = []
        with config.tqdm(desc='Downloading',
                         total=len(futures),
                         leave=config.pbar_leave,
                         disable=len(futures) == 1 or config.pbar_hide) as pbar:
            for f in as_completed(futures):
                pbar.update(1)
                try:
                    nl.append(f.result())
                except Exception as exc:
                    print(f'{futures[f]} generated an exception:', exc)

    return core.NeuronList(nl)


def _fetch_single_skeleton(img, **kwargs):
    """Fetch a single skeleton. Intended to be wrapped by ThreadPoolExecutor."""
    # Fetch the SWC table (oddly this seems to stall occasionally)
    r = requests.get(img['image_folder'] + '/volume.swc')
    swc = pd.read_csv(StringIO(r.content.decode()),
                      sep=' ', comment='#', header=None)
    swc.columns = ['node_id', 'label',
                   'x', 'y', 'z', 'radius',
                   'parent_id']
    # Turn into a TreeNeuron
    n = core.TreeNeuron(swc,
                        name=img['label'],
                        id=img['short_form'],
                        vfb_meta=img,
                        **kwargs)

    n._register_attr(name='template', value=img['template_anatomy']['label'])

    return n


def _query_images(short_forms, template, verbose=True):
    """Internal function to search for images."""
    short_forms = utils.make_iterable(short_forms).astype(str)

    # Split short forms into individuals, classes and others
    individuals = [s for s in short_forms if s.startswith('VFB')]
    classes = [s for s in short_forms if s.startswith('FBbt')]
    other = [s for s in short_forms if s not in individuals and s not in classes]

    # Get anything that matches the term(s)
    inds = []
    if individuals or other:
        inds += vc.neo_query_wrapper.get_anatomical_individual_TermInfo(short_forms=individuals + other)
    if classes or other:
        for hit in vc.neo_query_wrapper.get_TermInfo(short_forms=classes + other):
            inds += hit.get('anatomy_channel_image', [])

    # First get all images
    images = []
    for ind in inds:
        if 'term' in ind:
            data = ind['term']['core']
        elif 'anatomy' in ind:
            data = ind['anatomy']

        if not 'has_image' in data['types']:
            continue

        this_images = ind.get('channel_image', [])
        if isinstance(this_images, dict):
            this_images = [this_images]

        for im in this_images:
            images.append(im['image'])
            # Also carry some of the other info over
            images[-1]['label'] = data['label']
            images[-1]['short_form'] = data['short_form']

    # Gather all available templates
    templates = list(set([i['template_anatomy']['label'] for i in images]))

    if template:
        matches = [i for i in images if i['template_anatomy']['label'] == template]
    else:
        matches = images

    # Try giving some useful feedback why the search failed
    if verbose:
        if not matches:
            if not inds:
                logger.info(f'Searching for {short_forms} did not return any results')
            elif not images:
                logger.info(f'Matches for {short_forms} do not appear to have any associated '
                            'images')
            elif template and template not in templates:
                logger.info(f'Matches for {short_forms} do not have any images in template '
                            f'{template}. Try one of the following instead: '
                            f'{", ".join(templates)}')
        elif template:
            logger.info(f'Searching for {short_forms} returned {len(matches)} '
                        f'images in template {template}')
        else:
            logger.info(f'Searching for {short_forms} returned {len(matches)} '
                        f'images in {len(templates)} templates: '
                        f'{", ".join(templates)}')

    return matches



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.

from . import cave_utils

DATASTACK = "h01_c3_flat"
SERVER_ADDRESS = "https://global.brain-wire-test.org"
NUCLEUS_TABLE = "nucleus"


def get_cave_client(datastack=DATASTACK):
    """Get caveclient for H01 dataset."""
    client = cave_utils.get_cave_client(datastack, SERVER_ADDRESS)
    client.materialize.nucleus_table = NUCLEUS_TABLE
    return client


def fetch_neurons(
    x,
    *,
    lod=2,
    with_synapses=True,
    datastack=DATASTACK,
    materialization="auto",
    parallel=True,
    max_threads=4,
    **kwargs,
):
    """Fetch neuron meshes.

    Notes
    -----
    Synapses will be attached to the closest vertex on the mesh.

    Parameters
    ----------
    x :             str | int | list-like
                    Segment ID(s). Multiple IDs can be provided as list-like.
    lod :           int
                    Level of detail. Higher ``lod`` = coarser. This parameter
                    is ignored if the data source does not support multi-level
                    meshes.
    with_synapses : bool, optional
                    If True will also attach synapses as ``.connectors``.
    datastack :     str
                    Datastack to use. Default to "h01_c3_flat".
    materialization : "auto" | int
                    Which materialization version to use to look up somas and synapses
                    (if applicable). If "auto" (default) will try to find the most
                    recent version that contains the given root IDs. If an
                    integer is provided will use that version.
    parallel :      bool
                    If True, will use parallel threads to fetch data.
    max_threads :   int
                    Max number of parallel threads to use.
    **kwargs
                    Keyword arguments are passed through to the initialization
                    of the ``navis.MeshNeurons``.

    Returns
    -------
    navis.Neuronlist
                    Containing :class:`navis.MeshNeuron`.

    """
    client = get_cave_client(datastack)
    return cave_utils.fetch_neurons(
        x,
        lod=lod,
        with_synapses=with_synapses,
        client=client,
        parallel=parallel,
        max_threads=max_threads,
        materialization=materialization,
        **kwargs,
    )


def get_voxels(x, mip=0, bounds=None, datastack=DATASTACK):
    """Fetch voxels making a up given root ID.

    Parameters
    ----------
    x :             int
                    A single root ID.
    mip :           int
                    Scale at which to fetch voxels.
    bounds :        list, optional
                    Bounding box [xmin, xmax, ymin, ymax, zmin, zmax] in voxel
                    space. For example, the voxel resolution for mip 0
                    segmentation is 8 x 8 x 40 nm.
    datastack :     str
                    DATASTACK.

    Returns
    -------
    voxels :        (N, 3) np.ndarray
                    In voxel space according to `mip`.

    """
    return cave_utils.get_voxels(x, mip=mip, bounds=bounds, client=get_cave_client(datastack))



from copy import deepcopy

import navis

import pytest

from .common import with_igraph


def test_deepcopy():
    nrn = navis.core.BaseNeuron()
    deepcopy(nrn)


def test_from_swc(swc_source):
    n = navis.read_swc(swc_source)
    assert isinstance(n, navis.TreeNeuron)


@pytest.mark.parametrize("parallel", ["auto", True, 2, False])
def test_from_swc_multi(swc_source_multi, parallel):
    n = navis.read_swc(swc_source_multi, parallel=parallel)
    assert isinstance(n, navis.NeuronList)


@with_igraph
def test_from_gml():
    n = navis.example_neurons(n=1, source='gml')
    assert isinstance(n, navis.TreeNeuron)



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import scipy.optimize

import numpy as np
import pandas as pd

from typing import Optional

from .. import utils, config

logger = config.get_logger(__name__)


def bipartite_match(scores: pd.DataFrame,
                    one_to_one: bool = True,
                    threshold: Optional[float] = None,
                    labels: Optional[dict] = None,
                    known_matches: Optional[dict] = None,
                    duplicate_penalty: float = 0
                    ) -> dict:
    """Generate a bipartite matching that maximizes the scores.

    Parameters
    ----------
    scores :            pandas.Dataframe
                        *Similarity* scores (e.g. from NBLAST) to use for
                        bipartite matching. Index and columns will be used as
                        node IDs and must hence be unique.
    one_to_one :        bool
                        If True, we will enforce 1:1 matches. This will result
                        in unmatched nodes if `scores` is not quadratic or
                        if constraints (below) disallow certain matches. If
                        False, we will iteratively go through rounds of matching
                        in which we duplicate nodes opposite the axis with
                        missing nodes. A simple example to illustrate:

                        Let's assume we want to find a bipartite matching
                        of a 3x5 matrix without constraints. In a first 1:1
                        matching we can only determine 3 pairs leaving 2 nodes
                        unmatched. For those two nodes, we now duplicate their
                        best matches hence creating a 5x5 matrix and re-run
                        the matching. Without additional constraints this will
                        now return 5 pairs of one-to-many matches. With
                        additional constraints, we might have to undergo several
                        rounds of this.

                        Note this is not guaranteed to work if other
                        constraints (`threshold` or `restrict_labels`)
                        simply don't allow to match up every node.

    Constraints

    threshold :         int | float, optional
                        Similarities lower than this threshold will be
                        considered to weak for matching and ignored.
    labels :            dict, optional
                        A dictionary mapping node IDs (columns/indices) to
                        label(s). Only nodes that have the exact same set of
                        labels can be matched::

                          {'A': ['uPN', 'adPN'], 'B': ['uPN']}

                        In above example `A` can never match with `B`.

    known_matches :     dict, optional
                        Known matches can be enforced by providing a dictionary
                        mapping node U to V or vice versa::

                            {'A': 'B', 'B': 'C'}

    duplicate_penalty : float [0-1], optional
                        A score penalty for duplicated nodes. Each
                        time a node is duplicated for matching, this penalty
                        will be substracted. For example with
                        `duplicate_penalty = .1`::

                                               scores
                            original           1
                            first_duplicate    1 - 0.1 = 0.9
                            second_duplicate   0.9 - 0.1 = 0.8

                        Leave at `0` for no penalties.

    Returns
    -------
    pandas.DataFrame
                        Ummatched nodes will show up without a counterpart
                        or score.

    Raises
    ------
    ValueError
                        `scipy.optimize.linear_sum_assignment` raises an
                        "cost matrix is infeasible" exception when constraints
                        make a matching impossible. Try releasing some
                        constraints.

    """
    assert isinstance(scores, pd.DataFrame)
    assert one_to_one in [False, None, True]

    if not duplicate_penalty:
        duplicate_penalty = 0

    # Make a copy
    scores = scores.copy()

    # Turn such that there are more rows than columns
    if scores.shape[0] < scores.shape[1]:
        scores = scores.T

    # For some odd reason we get funny results if index and column are not str
    scores.index = scores.index.astype(str)
    scores.columns = scores.columns.astype(str)

    # Make sure columns and indices are unique
    if any(scores.index.duplicated()) or any(scores.columns.duplicated()):
        raise ValueError('Indices and columns in scores DataFrame must be unique.')

    # Turn into edges
    ix_name = scores.index.name if scores.index.name else 'index'
    edges = scores.T.unstack().reset_index()
    edges.columns = ['U', 'V', 'score']

    # Set distance of below threshold scores to an insanely high value
    # Must not use infinite - otherwise scipy's optimization throws hissy fits
    if not isinstance(threshold, type(None)):
        # Convert threshold to distances
        edges.loc[edges.score < threshold, 'score'] = -1e10  # can't use inf here unfortunately

    # Drop edges between nodes with labels that don't match
    if isinstance(labels, dict):
        # Make sure items are sets
        rl = {k: set(utils.make_iterable(v)) for k, v in labels.items()}

        # Do not change the below! This comparison looks clunky but any other
        # way we run into issues where all of the sudden `np.nan == np.nan = False`
        ul = edges.U.map(lambda x: rl.get(x, None)).values
        vl = edges.V.map(lambda x: rl.get(x, None)).values
        edges.loc[ul != vl, 'score'] = -1e10  # can't use inf here unfortunately

    # Enforce premade matches by adding edges with weight infinity
    if isinstance(known_matches, dict):
        # This is rather slow atm and we should try speeding this up
        match_sets = [{str(k), str(v)} for k, v in known_matches.items()]
        is_match = edges[['U', 'V']].apply(set, axis=1).apply(lambda x: x in match_sets)
        edges.loc[is_match, 'score'] = np.inf

    # Turn back into matrix
    scores_final = edges.pivot(index='U', columns='V', values='score')

    # Note: it looks like we're allowed to have rows with only -infinity or
    # columns with only -infinity values but not both. Otherwise scipy will
    # complain.

    # Make matches
    left_ix, right_ix = scipy.optimize.linear_sum_assignment(scores_final, maximize=True)

    # Drop matches that have effectively minus infinity scores
    is_above = scores_final.values[left_ix, right_ix] > -1e10
    left_ix, right_ix = left_ix[is_above], right_ix[is_above]


    # Convert to row/column names
    left_id = scores_final.index[left_ix]
    right_id = scores_final.columns[right_ix]

    # Check if we need to rematch -> note we are only checking indices (axis 0)
    missing = np.array(list(set(scores.index) - set(left_id)))

    # We can only rematch neurons that have a possible match to being with
    missing = missing[scores_final.loc[missing].max(axis=1) > -1e10]

    if not one_to_one and len(missing) > 0:
        while len(missing) > 0:
            # Find the best match for the missing nodes
            best_match = np.argmax(scores_final.loc[missing].values, axis=1)

            # Add virtual clones of these best matches
            clones = scores_final.iloc[:, best_match] - duplicate_penalty
            scores_final = pd.concat([scores_final, clones], axis=1)

            # Make pairs again
            left_ix, right_ix = scipy.optimize.linear_sum_assignment(scores_final, maximize=True)

            # Drop matches that have effectively minus infinity scores
            is_above = scores_final.values[left_ix, right_ix] > 1e-10
            left_ix, right_ix = left_ix[is_above], right_ix[is_above]

            # Convert to row/column names
            left_id = scores_final.index[left_ix]
            right_id = scores_final.columns[right_ix]

            # Check if we need to rematch
            # -> note we are only checking columns now (axis 0)
            missing = np.array(list(set(scores_final.columns) - set(right_id)))

            # We can only rematch neurons that have a possible match to begin with
            missing = missing[scores_final[missing].max(axis=0) > 1e-10]

            # Transpose matrix
            scores_final = scores_final.T

        # One final transpose to get back in original configuration
        scores_final = scores_final.T

    matches = pd.DataFrame()
    matches['U'] = left_id
    matches['V'] = right_id

    # Add scores
    matches['score'] = scores_final.values[left_ix, right_ix]

    # Add neurons that have not been matches
    matched = np.unique(matches[['U', 'V']].values.flatten())
    U_miss = pd.DataFrame([])
    U_miss['U'] = scores_final.index[~scores_final.index.isin(matched)]
    U_miss['V'] = None
    U_miss['score'] = None

    V_miss = pd.DataFrame([])
    V_miss['U'] = None
    V_miss['V'] = scores_final.columns[~scores_final.columns.isin(matched)]
    V_miss['score'] = None

    if not U_miss.empty or not V_miss.empty:
        logger.info(f'{U_miss.shape[0] + V_miss.shape[0]} neurons could '
                           'not be matched under the given constraints.')

    return pd.concat([matches, U_miss, V_miss], axis=0).reset_index(drop=True)



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

from .iterables import make_iterable, make_non_iterable, is_iterable, multi_split
from .misc import (is_jupyter, set_loggers, set_pbars, unpack_neurons,
                   set_default_connector_colors, parse_objects,
                   is_url, make_url, make_volume, sizeof_fmt,
                   round_smart, is_blender, check_vispy)
from .validate import validate_options, validate_table
from .eval import (eval_node_ids, eval_neurons, eval_id, eval_conditions,
                   is_mesh, is_numeric, eval_param)
from .exceptions import (ConstructionError, VolumeError, CMTKError)
from .cv import (patch_cloudvolume)
from .decorators import (meshneuron_skeleton, map_neuronlist_df, map_neuronlist,
                         lock_neuron)

try:
    import navis_fastcore as fastcore
except ModuleNotFoundError:
    fastcore = None

__all__ = ['set_loggers', 'set_pbars', 'set_default_connector_colors',
           'patch_cloudvolume']



from typing import Iterable, Optional, NamedTuple
from ..core import TreeNeuron
import networkx as nx
import pandas as pd
import numpy as np
from ..config import get_logger

logger = get_logger(__name__)

OTHER = "__OTHER__"


class Edge(NamedTuple):
    connector_id: int
    source_name: str
    target_name: str
    source_node: Optional[int]
    target_node: Optional[int]


class NeuronConnector:
    """Class which creates a connectivity graph from a set of neurons.

    Connectivity is determined by shared IDs in the `connectors` table.

    Add neurons with the `add_neuron` and `add_neurons` methods.
    Alternatively, supply an iterable of neurons in the constructor.
    Neurons must have unique names.

    See the `to_(multi)digraph` method for output.
    """

    def __init__(self, nrns: Optional[Iterable[TreeNeuron]] = None) -> None:
        self.neurons = dict()
        self.connector_xyz = dict()
        # connectors and the treenodes presynaptic to them
        self.conn_inputs = dict()
        # connectors and the treenodes postsynaptic to them
        self.conn_outputs = dict()

        if nrns is not None:
            self.add_neurons(nrns)

    def __len__(self) -> int:
        return len(self.neurons)

    def add_neurons(self, nrns: Iterable[TreeNeuron]):
        """Add several neurons to the connector.

        All neurons must have unique names.

        Parameters
        ----------
        nrns : Iterable[TreeNeuron]

        Returns
        -------
        Modified connector.
        """
        for nrn in nrns:
            self.add_neuron(nrn)
        return self

    def add_neuron(self, nrn: TreeNeuron):
        """Add a single neuron to the connector.

        All neurons must have unique names.

        Parameters
        ----------
        nrn : TreeNeuron

        Returns
        -------
        Modified connector.
        """
        if nrn.name in self.neurons:
            logger.warning(
                "Neuron with name %s has already been added to NeuronConnector. "
                "These will occupy the same node in the graph, "
                "but have connectors from both.",
                nrn.name
            )

        self.neurons[nrn.name] = nrn
        if nrn.connectors is None:
            logger.warning("Neuron with name %s has no connector information", nrn.name)
            return self

        for row in nrn.connectors.itertuples():
            # connector_id, node_id, x, y, z, is_input
            self.connector_xyz[row.connector_id] = (row.x, row.y, row.z)
            if row.type == 1:
                self.conn_outputs.setdefault(row.connector_id, []).append((nrn.name, row.node_id))
            elif row.type == 0:
                if row.connector_id in self.conn_inputs:
                    logger.warning(
                        "Connector with ID %s has multiple inputs: "
                        "connector tables are probably inconsistent",
                        row.connector_id
                    )
                self.conn_inputs[row.connector_id] = (nrn.name, row.node_id)

        return self

    def edges(self, include_other=True) -> Iterable[Edge]:
        """Iterate through all synapse edges.

        Parameters
        ----------
        include_other : bool, optional
            Include edges for which only one partner is known, by default True.
            If included, the name of the unknown partner will be `"__OTHER__"`,
            and the treenode ID will be None.

        Yields
        ------
        tuple[int, str, str, int, int]
            Connector ID, source name, target name, source treenode, target treenode.
        """
        for conn_id in set(self.conn_inputs).union(self.conn_outputs):
            src, src_node = self.conn_inputs.get(conn_id, (OTHER, None))
            if src_node is None and not include_other:
                continue
            for tgt, tgt_node in self.conn_outputs.get(conn_id, [(OTHER, None)]):
                if tgt_node is None and not include_other:
                    continue
                yield Edge(conn_id, src, tgt, src_node, tgt_node)

    def to_adjacency(self, include_other=True) -> pd.DataFrame:
        """Create an adjacency matrix of neuron connectivity.

        Parameters
        ----------
        include_other : bool, optional
            Whether to include a node called `"__OTHER__"`,
            which represents all unknown partners.
            By default True.
            This can be helpful when calculating a neuron's input fraction,
            but cannot be used for output fractions if synapses are polyadic.

        Returns
        -------
        pandas.DataFrame
            Row index is source neuron name,
            column index is target neuron name,
            cells are the number of synapses from source to target.
        """
        index = list(self.neurons)
        if include_other:
            index.append(OTHER)
        data = np.zeros((len(index), len(index)), np.uint64)
        df = pd.DataFrame(data, index, index)
        for _, src, tgt, _, _ in self.edges(include_other):
            df.loc[src, tgt] += 1

        return df

    def to_digraph(self, include_other=True) -> nx.DiGraph:
        """Create a graph of neuron connectivity.

        Parameters
        ----------
        include_other : bool, optional
            Whether to include a node called `"__OTHER__"`,
            which represents all unknown partners.
            By default True.
            This can be helpful when calculating a neuron's input fraction,
            but cannot be used for output fractions if synapses are polyadic.

        Returns
        -------
        nx.DiGraph
            The graph has data `{"connector_xyz": {connector_id: (x, y, z), ...}}`.
            The nodes have data `{"neuron": tree_neuron}`.
            The edges have data `{"connectors": data_frame, "weight": n_connectors}`,
            where the connectors data frame has columns
            "connector_id", "pre_node", "post_node".
        """
        g = nx.DiGraph()
        g.add_nodes_from((k, {"neuron": v}) for k, v in self.neurons.items())
        if include_other:
            g.add_node(OTHER, neuron=None)

        g.graph["connector_xyz"] = self.connector_xyz
        headers = {
            "connector_id": pd.UInt64Dtype(),
            "pre_node": pd.UInt64Dtype(),
            "post_node": pd.UInt64Dtype(),
        }
        edges = dict()
        for conn_id, src, tgt, src_node, tgt_node in self.edges(include_other):
            edges.setdefault((src, tgt), []).append([conn_id, src_node, tgt_node])

        for (src, tgt), rows in edges.items():
            df_tmp = pd.DataFrame(rows, columns=list(headers), dtype=object)
            df = df_tmp.astype(headers, copy=False)
            g.add_edge(src, tgt, connectors=df, weight=len(df))

        return g

    def to_multidigraph(self, include_other=True) -> nx.MultiDiGraph:
        """Create a graph of neuron connectivity where each synapse is an edge.

        Parameters
        ----------
        include_other : bool, optional
            Whether to include a node called `"__OTHER__"`,
            which represents all unknown partners.
            By default True.
            This can be helpful when calculating a neuron's input fraction,
            but cannot be used for output fractions if synapses are polyadic.

        Returns
        -------
        nx.MultiDiGraph
            The nodes have data `{"neuron": tree_neuron}`.
            The edges have data
            `{"pre_node": presyn_treenode_id, "post_node": postsyn_treenode_id, "xyz": connector_location, "connector_id": conn_id}`.
        """
        g = nx.MultiDiGraph()
        g.add_nodes_from((k, {"neuron": v}) for k, v in self.neurons.items())
        if include_other:
            g.add_node(OTHER, neuron=None)

        for conn_id, src, tgt, src_node, tgt_node in self.edges(include_other):
            g.add_edge(
                src,
                tgt,
                pre_node=src_node,
                post_node=tgt_node,
                xyz=self.connector_xyz[conn_id],
                connector_id=conn_id,
            )

        return g



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

"""Module containing base classes for BLASTING."""

import numpy as np
import pandas as pd

from abc import ABC, abstractmethod
from typing import Union, List

from .. import utils, config

INT_DTYPES = {16: np.int16, 32: np.int32, 64: np.int64, None: None,
              'single': np.int32, 'double': np.int64}
FLOAT_DTYPES = {16: np.float16, 32: np.float32, 64: np.float64, None: None,
                'single': np.float32, 'double': np.float64}

logger = config.logger

NestedIndices = Union[int, List['NestedIndices']]


class Blaster(ABC):
    """Base class for blasting."""

    def __init__(self, dtype=np.float64, progress=True):
        """Initialize class."""
        self.dtype = dtype
        self.progress = progress
        self.desc = "Blasting"
        self.self_hits = []
        self.neurons = []
        self.ids = []

    def __len__(self):
        return len(self.neurons)

    @abstractmethod
    def append(self, neurons) -> NestedIndices:
        """Append neurons."""
        pass

    @abstractmethod
    def calc_self_hit(self, neurons):
        """Non-normalized value for self hit."""
        pass

    @abstractmethod
    def single_query_target(self, q_idx, t_idx, scores='forward'):
        """Query single target against single target."""
        pass

    @property
    def dtype(self):
        """Data type used for scores."""
        return self._dtype

    @dtype.setter
    def dtype(self, dtype):
        try:
            self._dtype = np.dtype(dtype)
        except TypeError:
            try:
                self._dtype = FLOAT_DTYPES[dtype]
            except KeyError:
                raise ValueError(
                    f'Unknown precision/dtype {dtype}. Expected on of the following: 16, 32 or 64 (default)'
                )

    def pair_query_target(self, pairs, scores='forward'):
        """BLAST multiple pairs.

        Parameters
        ----------
        pairs :             tuples
                            Tuples of (query_ix, target_ix) to query.
        scores :            "forward" | "mean" | "min" | "max" | "both"
                            Which scores to return.

        """
        # See `multi_query_target` for explanation on progress bars
        scr = []
        for p in config.tqdm_classic(pairs,
                             desc=f'{self.desc} pairs',
                             leave=False,
                             position=getattr(self,
                                              'pbar_position',
                                              None),
                             disable=not self.progress):
            scr.append(self.single_query_target(p[0], p[1], scores=scores))

        return scr

    def multi_query_target(self, q_idx, t_idx, scores='forward'):
        """BLAST multiple queries against multiple targets.

        Parameters
        ----------
        q_idx,t_idx :       iterable
                            Iterable of query/target neuron indices to BLAST.
        scores :            "forward" | "mean" | "min" | "max" | "both"
                            Which scores to return.

        """
        shape = (len(q_idx), len(t_idx)) if scores != 'both' else (len(q_idx), len(t_idx), 2)
        res = np.empty(shape, dtype=self.dtype)
        for i, q in enumerate(config.tqdm(q_idx,
                                          desc=self.desc,
                                          leave=False,
                                          position=getattr(self,
                                                           'pbar_position',
                                                           None),
                                          disable=not self.progress)):
            for k, t in enumerate(t_idx):
                res[i, k] = self.single_query_target(q, t, scores=scores)

        # Generate results
        if res.ndim == 2:
            res = pd.DataFrame(res)
            res.columns = [self.ids[t] for t in t_idx]
            res.index = [self.ids[q] for q in q_idx]
            res.index.name = 'query'
            res.columns.name = 'target'
        else:
            # For scores='both' we will create a DataFrame with multi-index
            ix = pd.MultiIndex.from_product([[self.ids[q] for q in q_idx],
                                             ['forward', 'reverse']],
                                            names=["query", "score"])
            res = pd.DataFrame(np.hstack((res[:, :, 0],
                                          res[:, :, 1])).reshape(len(q_idx) * 2,
                                                                 len(t_idx)),
                               index=ix,
                               columns=[self.ids[t] for t in t_idx])
            res.columns.name = 'target'

        return res

    def all_by_all(self, scores='forward'):
        """BLAST all-by-all neurons."""
        res = self.multi_query_target(range(len(self.neurons)),
                                      range(len(self.neurons)),
                                      scores='forward')

        # For all-by-all BLAST we can get the mean score by
        # transposing the scores
        if scores == 'mean':
            res = (res + res.T) / 2
        elif scores == 'min':
            res.loc[:, :] = np.dstack((res, res.T)).min(axis=2)
        elif scores == 'max':
            res.loc[:, :] = np.dstack((res, res.T)).max(axis=2)
        elif scores == 'both':
            ix = pd.MultiIndex.from_product([res.index, ['forward', 'reverse']],
                                            names=["query", "score"])
            res = pd.DataFrame(np.hstack((res[:, :, 0],
                                          res[:, :, 1])).reshape(res.shape[0] * 2,
                                                                 res.shape[1]),
                               index=ix,
                               columns=res.columns)

        return res

    def __len__(self):
        return len(self.neurons)



import numpy as np

from scipy import ndimage
from tqdm.auto import trange

from .. import core, config
from .base import TransformSequence, BaseTransform
from .h5reg import H5transform
from . import registry

logger = config.get_logger(__name__)


class ImageXformer:
    """Class to render images from a source to a target space.

    You should be able to gather most parameters from a TemplateBrain object.

    Parameters
    ----------
    transform :     Transform | Sequence
                    Transform or sequence of transforms to go
                    *from target to source* template space.
    target_dims :   (x, y, z) tuple
                    Dimensions of the target space, i.e. how many voxels in each
                    axis.
    target_spacing : (x, y z) tuple
                    Size of voxels in the target space.
    source_spacing : (x, y, z) tuple
                    Size of voxels in the source space.
    target_offset : (x, y, z) tuple, optional
                    Offset in physical units of the target space. This will
                    typically be (0, 0, 0).
    source_offset : (x, y, z) tuple, optional
                    Offset in physical units of the source space. This will
                    typically be (0, 0, 0).
    progress :      bool
                    Whether to show a progress bar for the rendering.

    """
    def __init__(
        self,
        transform,
        target_dims,
        target_spacing,
        source_spacing,
        target_offset=None,
        source_offset=None,
        progress: bool = True,
    ):
        if isinstance(transform, BaseTransform):
            transform = TransformSequence(transform)

        if isinstance(transform, list):
            transform = TransformSequence(*transform)

        self.transform = transform
        self.target_dims = target_dims
        self.target_spacing = target_spacing
        self.target_offset = target_offset
        self.source_spacing = source_spacing
        self.source_offset = source_offset
        self.progress = progress
        self.stepsize = int(1e6)
        self.interpolation_order = 1

        # See if we can pre-cache the transform(s)
        if self.progress:
            logger.info('Caching transform(s).')
        if isinstance(self.transform, H5transform):
            self.transform.full_ingest()
        elif isinstance(transform, TransformSequence):
            for t in transform.transforms:
                if isinstance(t, H5transform):
                    t.full_ingest()

    def render(self, image):
        """Render an image into the target space.

        Parameters
        ----------
        image : VoxelNeuron | (M, N, K) numpy array
                Image in source space to transform.

        Returns
        -------
        VoxelNeuron
                Image in target space.

        """
        if isinstance(image, core.VoxelNeuron):
            image = image.grid

        # Generate the coordinates for each pixel in the target space
        XX, YY, ZZ = np.meshgrid(
            range(self.target_dims[0]),
            range(self.target_dims[1]),
            range(self.target_dims[2]),
            indexing="ij",
        )
        # Generate a (3, M, N, K) grid
        ix_grid = np.array([XX, YY, ZZ])

        # Convert grid into (N * N * K, 3) voxel array (this is slow)
        ix_array_target = ix_grid.T.reshape(-1, 3)
        del ix_grid  # we don't need it anymore

        # Prepare the image to render onto
        img_xf = np.zeros(self.target_dims, dtype=image.dtype)

        # Render in batches to keep memory consumption down
        bar_format = "{desc}{percentage:3.0f}%|{bar}|[{elapsed}<{remaining}]"
        for i in trange(
            0,
            ix_array_target.shape[0],
            self.stepsize,
            desc="Rendering",
            disable=not self.progress,
            bar_format=bar_format,
        ):
            # Convert indices from voxel to physical coordinates we can transform
            coo_array_target = ix_array_target[i : i + self.stepsize] * self.target_spacing

            # Add physical offset if applicable
            if self.target_offset is not None:
                coo_array_target += self.target_offset

            # Project these coordinates from target to source space
            # This step is the the bottleneck since we are (potentially) xforming
            # millions of coordinates
            current_level = int(logger.level)
            try:
                logger.setLevel("ERROR")
                # Convert this batch of physical coordinates into physical source space
                coo_array_source = self.transform.xform(
                    coo_array_target, affine_fallback=True
                )
            except BaseException:
                raise
            finally:
                logger.setLevel(current_level)

            # Convert physical coordinates into voxels (note that we are NOT rounding here)
            ix_array_source = coo_array_source / self.source_spacing

            if self.source_offset is not None:
                ix_array_source += self.source_offset

            # Use target->source index mapping to interpolate the image
            # order=1 means linear interpolation (much faster)
            img_xf[
                ix_array_target[i : i + self.stepsize, 0],
                ix_array_target[i : i + self.stepsize, 1],
                ix_array_target[i : i + self.stepsize, 2],
            ] = ndimage.map_coordinates(image, ix_array_source.T, order=self.interpolation_order)

        return core.VoxelNeuron(img_xf, offset=self.target_offset, units=self.target_spacing)


def xform_image(img, source, target, progress=True):
    """Experimental function to render image into a target space.

    Parameters
    ----------
    img :       VoxelNeuron | (M, N, K) numpy array
                Image to transform.
    source :    str
                The source template space.
    target :    str
                The target template space.
    progress :  bool
                Whether to show a progress bar.

    Returns
    -------
    VoxelNeuron
                Image in target space.

    """
    if isinstance(img, core.VoxelNeuron):
        img = img.grid
    elif not isinstance(img, np.ndarray):
        raise TypeError(f'Expected VoxelNeuron or numpy array, got "{type(img)}"')
    elif img.ndim != 3:
        raise ValueError('Image must be 3D.')

    # Get the transform from target to source
    transform = registry.find_bridging_path(target, source)

    # We need info on the source and target spaces
    source = registry.find_template(source)
    target = registry.find_template(target)

    # Generate the ImageXformer
    xformer = ImageXformer(
        transform,
        target.dims,
        target.voxdims,
        source.voxdims,
        target_offset=np.asarray(target.boundingbox).reshape(3, 2)[:, 0],
        source_offset=np.asarray(source.boundingbox).reshape(3, 2)[:, 0],
        progress=progress
    )

    return xformer.xform(img)



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import os
import io

import trimesh as tm

from typing import Union, Iterable, Optional, Dict, Any
from typing_extensions import Literal
from urllib3 import HTTPResponse

from .. import config, utils, core
from . import base

# Set up logging
logger = config.get_logger(__name__)

# Mesh files can have all sort of extensions
DEFAULT_FMT = "{name}.{file_ext}"

# Mesh extensions supported by trimesh
MESH_LOAD_EXT = tuple(tm.exchange.load.mesh_loaders.keys())
MESH_WRITE_EXT = tuple(tm.exchange.export._mesh_exporters.keys())


class MeshReader(base.BaseReader):
    def __init__(
        self,
        output: str,
        fmt: str = DEFAULT_FMT,
        attrs: Optional[Dict[str, Any]] = None,
        errors: str = "raise",
    ):
        super().__init__(
            fmt=fmt,
            attrs=attrs,
            file_ext=MESH_LOAD_EXT,
            name_fallback="MESH",
            read_binary=True,
            errors=errors,
        )
        self.output = output

    def format_output(self, x):
        # This function replaces the BaseReader.format_output()
        # This is to avoid trying to convert multiple (image, header) to NeuronList
        if self.output == "trimesh":
            return x
        elif x:
            return core.NeuronList(x)
        else:
            return core.NeuronList([])

    @base.handle_errors
    def read_buffer(
        self, f, attrs: Optional[Dict[str, Any]] = None
    ) -> Union[tm.Trimesh, "core.Volume", "core.MeshNeuron"]:
        """Read buffer into mesh.

        Parameters
        ----------
        f :         IO
                    Readable buffer (must be bytes).
        attrs :     dict | None
                    Arbitrary attributes to include in the neurons.

        Returns
        -------
        Trimesh | MeshNeuron | Volume

        """
        if isinstance(f, HTTPResponse):
            f = io.StringIO(f.content)

        if isinstance(f, bytes):
            f = io.BytesIO(f)

        # We need to tell trimesh what file type we are reading
        if "file" not in attrs:
            raise KeyError(
                f'Unable to parse file type. "file" not in attributes: {attrs}'
            )

        file_type = attrs["file"].split(".")[-1]

        mesh = tm.load_mesh(f, file_type=file_type)

        if self.output == "trimesh":
            return mesh
        elif self.output == "volume":
            return core.Volume(mesh.vertices, mesh.faces, **attrs)

        # Turn into a MeshNeuron
        n = core.MeshNeuron(mesh)

        # Try adding properties one-by-one. If one fails, we'll keep track of it
        # in the `.meta` attribute
        meta = {}
        for k, v in attrs.items():
            try:
                n._register_attr(k, v)
            except (AttributeError, ValueError, TypeError):
                meta[k] = v

        if meta:
            n.meta = meta

        return n


def read_mesh(
    f: Union[str, Iterable],
    include_subdirs: bool = False,
    parallel: Union[bool, int] = "auto",
    output: Union[Literal["neuron"], Literal["volume"], Literal["trimesh"]] = "neuron",
    errors: Literal["raise", "log", "ignore"] = "raise",
    limit: Optional[int] = None,
    fmt: str = "{name}.",
    **kwargs,
) -> "core.NeuronObject":
    """Load mesh file into Neuron/List.

    This is a thin wrapper around `trimesh.load_mesh` which supports most
    commonly used formats (obj, ply, stl, etc.).

    Parameters
    ----------
    f :                 str | iterable
                        Filename(s) or folder. If folder should include file
                        extension (e.g. `my/dir/*.ply`) otherwise all
                        mesh files in the folder will be read.
    include_subdirs :   bool, optional
                        If True and `f` is a folder, will also search
                        subdirectories for meshes.
    parallel :          "auto" | bool | int,
                        Defaults to `auto` which means only use parallel
                        processing if more than 100 mesh files are imported.
                        Spawning and joining processes causes overhead and is
                        considerably slower for imports of small numbers of
                        neurons. Integer will be interpreted as the number of
                        cores (otherwise defaults to `os.cpu_count() - 2`).
    output :            "neuron" | "volume" | "trimesh"
                        Determines function's output - see `Returns`.
    errors :            "raise" | "log" | "ignore"
                        If "log" or "ignore", errors will not be raised and the
                        mesh will be skipped. Can result in empty output.
    limit :             int | str | slice | list, optional
                        When reading from a folder or archive you can use this parameter to
                        restrict the which files read:
                         - if an integer, will read only the first `limit` mesh files
                           (useful to get a sample from a large library of meshes)
                         - if a string, will interpret it as filename (regex) pattern
                           and only read files that match the pattern; e.g. `limit='.*_R.*'`
                           will only read files that contain `_R` in their filename
                         - if a slice (e.g. `slice(10, 20)`) will read only the files in
                           that range
                         - a list is expected to be a list of filenames to read from
                           the folder/archive
    **kwargs
                        Keyword arguments passed to [`navis.MeshNeuron`][]
                        or [`navis.Volume`][]. You can use this to e.g.
                        set the units on the neurons.

    Returns
    -------
    MeshNeuron
                        If `output="neuron"` (default).
    Volume
                        If `output="volume"`.
    Trimesh
                        If `output="trimesh"`.
    NeuronList
                        If `output="neuron"` and import has multiple meshes
                        will return NeuronList of MeshNeurons.
    list
                        If `output!="neuron"` and import has multiple meshes
                        will return list of Volumes or Trimesh.

    See Also
    --------
    [`navis.read_precomputed`][]
                        Read meshes and skeletons from Neuroglancer's precomputed format.

    Examples
    --------

    Read a single file into [`navis.MeshNeuron`][]:

    >>> m = navis.read_mesh('mesh.obj')                         # doctest: +SKIP

    Read all e.g. .obj files in a directory:

    >>> nl = navis.read_mesh('/some/directory/*.obj')           # doctest: +SKIP

    Sample first 50 files in folder:

    >>> nl = navis.read_mesh('/some/directory/*.obj', limit=50) # doctest: +SKIP

    Read single file into [`navis.Volume`][]:

    >>> nl = navis.read_mesh('mesh.obj', output='volume')       # doctest: +SKIP

    """
    utils.eval_param(
        output, name="output", allowed_values=("neuron", "volume", "trimesh")
    )

    reader = MeshReader(fmt=fmt, output=output, errors=errors, attrs=kwargs)
    return reader.read_any(f, include_subdirs, parallel, limit=limit)


def write_mesh(
    x: Union["core.NeuronList", "core.MeshNeuron", "core.Volume", "tm.Trimesh"],
    filepath: Optional[str] = None,
    filetype: str = None,
) -> None:
    """Export meshes (MeshNeurons, Volumes, Trimeshes) to disk.

    Under the hood this is using trimesh to export meshes.

    Parameters
    ----------
    x :                 MeshNeuron | Volume | Trimesh | NeuronList
                        If multiple objects, will generate a file for each
                        neuron (see also `filepath`).
    filepath :          None | str | list, optional
                        If `None`, will return byte string or list of
                        thereof. If filepath will save to this file. If path
                        will save neuron(s) in that path using `{x.id}`
                        as filename(s). If list, input must be NeuronList and
                        a filepath must be provided for each neuron.
    filetype :          stl | ply | obj, optional
                        If `filepath` does not include the file extension,
                        you need to provide it as `filetype`.

    Returns
    -------
    None
                        If filepath is not `None`.
    bytes
                        If filepath is `None`.

    See Also
    --------
    [`navis.read_mesh`][]
                        Import neurons.
    [`navis.write_precomputed`][]
                        Write meshes to Neuroglancer's precomputed format.

    Examples
    --------

    Write `MeshNeurons` to folder:

    >>> import navis
    >>> nl = navis.example_neurons(3, kind='mesh')
    >>> navis.write_mesh(nl, tmp_dir, filetype='obj')

    Specify the filenames:

    >>> import navis
    >>> nl = navis.example_neurons(3, kind='mesh')
    >>> navis.write_mesh(nl, tmp_dir / '{neuron.name}.obj')

    Write directly to zip archive:

    >>> import navis
    >>> nl = navis.example_neurons(3, kind='mesh')
    >>> navis.write_mesh(nl, tmp_dir / 'meshes.zip', filetype='obj')

    """
    if filetype is not None:
        utils.eval_param(filetype, name="filetype", allowed_values=MESH_WRITE_EXT)
    else:
        # See if we can get filetype from filepath
        if filepath is not None:
            for f in MESH_WRITE_EXT:
                if str(filepath).endswith(f".{f}"):
                    filetype = f
                    break

        if not filetype:
            raise ValueError(
                "Must provide mesh type either explicitly via "
                "`filetype` variable or implicitly via the "
                "file extension in `filepath`"
            )

    writer = base.Writer(_write_mesh, ext=f".{filetype}")

    return writer.write_any(x, filepath=filepath)


def _write_mesh(
    x: Union["core.MeshNeuron", "core.Volume", "tm.Trimesh"],
    filepath: Optional[str] = None,
) -> None:
    """Write single mesh to disk."""
    if filepath and os.path.isdir(filepath):
        if isinstance(x, core.MeshNeuron):
            if not x.id:
                raise ValueError(
                    "Neuron(s) must have an ID when destination " "is a folder"
                )
            filepath = os.path.join(filepath, f"{x.id}")
        elif isinstance(x, core.Volume):
            filepath = os.path.join(filepath, f"{x.name}")
        else:
            raise ValueError(f"Unable to generate filename for {type(x)}")

    if isinstance(x, core.MeshNeuron):
        mesh = x.trimesh
    elif isinstance(x, tm.Trimesh):
        mesh = x
    else:
        raise TypeError(f'Unable to write data of type "{type(x)}"')

    # Write to disk (will only return content if filename is None)
    content = mesh.export(filepath)

    if not filepath:
        return content



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import os

import multiprocessing as mp
import networkx as nx
import numpy as np
import pandas as pd
import trimesh as tm

from typing import Union
from scipy import ndimage, stats
from tqdm.auto import tqdm

try:
    import skimage
    from skimage import measure
except ModuleNotFoundError:
    skimage = None

from .. import core, config, intersection, graph, morpho


logger = config.get_logger(__name__)


def fix_mesh(mesh: Union[tm.Trimesh, 'core.MeshNeuron'],
             fill_holes: bool = False,
             remove_fragments: bool = False,
             inplace: bool = False):
    """Try to fix some common problems with mesh.

     1. Remove infinite values
     2. Merge duplicate vertices
     3. Remove duplicate and degenerate faces
     4. Fix normals
     5. Remove unreference vertices
     6. Remove disconnected fragments (Optional)
     7. Fill holes (Optional)

    Parameters
    ----------
    mesh :              trimesh.Trimesh | navis.MeshNeuron
    fill_holes :        bool
                        If True will try to fix holes in the mesh.
    remove_fragments :  False | int
                        If a number is given, will iterate over the mesh's
                        connected components and remove those consisting of less
                        than the given number of vertices. For example,
                        `remove_fragments=5` will drop parts of the mesh
                        that consist of five or less connected vertices.
    inplace :           bool
                        If True, will perform fixes on the input mesh. If False,
                        will make a copy and leave the original untouched.

    Returns
    -------
    fixed object :      trimesh.Trimesh or navis.MeshNeuron

    """
    if not inplace:
        mesh = mesh.copy()

    if isinstance(mesh, core.MeshNeuron):
        m = mesh.trimesh
    else:
        m = mesh

    assert isinstance(m, tm.Trimesh)

    if remove_fragments:
        to_drop = []
        for c in nx.connected_components(m.vertex_adjacency_graph):
            if len(c) <= remove_fragments:
                to_drop += list(c)

        # Remove dropped vertices
        remove = np.isin(np.arange(m.vertices.shape[0]), to_drop)
        m.update_vertices(~remove)

    if fill_holes:
        m.fill_holes()

    m.remove_infinite_values()
    m.merge_vertices()
    m.remove_duplicate_faces()
    m.remove_degenerate_faces()
    m.fix_normals()
    m.remove_unreferenced_vertices()

    # If we started with a MeshNeuron, map back the verts/faces
    if isinstance(mesh, core.MeshNeuron):
        mesh.vertices, mesh.faces = m.vertices, m.faces
        mesh._clear_temp_attr()

    return mesh


def smooth_mesh_trimesh(x, iterations=5, L=0.5, inplace=False):
    """Smooth mesh using Trimesh's Laplacian smoothing.

    Parameters
    ----------
    x :             MeshNeuron | Volume | Trimesh
                    Mesh object to simplify.
    iterations :    int
                    Round of smoothing to apply.
    L :             float [0-1]
                    Diffusion speed constant lambda. Larger = more aggressive
                    smoothing.
    inplace :       bool
                    If True, will perform simplication on `x`. If False, will
                    simplify and return a copy.

    Returns
    -------
    simp
                Simplified mesh object.

    """
    if L > 1 or L < 0:
        raise ValueError(f'`L` (lambda) must be between 0 and 1, got "{L}"')

    if isinstance(x, core.MeshNeuron):
        mesh = x.trimesh.copy()
    elif isinstance(x, core.Volume):
        mesh = tm.Trimesh(x.vertices, x.faces)
    elif isinstance(x, tm.Trimesh):
        mesh = x.copy()
    else:
        raise TypeError('Expected MeshNeuron, Volume or trimesh.Trimesh, '
                        f'got "{type(x)}"')

    assert isinstance(mesh, tm.Trimesh)

    # Smooth mesh
    # This always happens in place, hence we made a copy earlier
    tm.smoothing.filter_laplacian(mesh, lamb=L, iterations=iterations)

    if not inplace:
        x = x.copy()

    x.vertices = mesh.vertices
    x.faces = mesh.faces

    return x


def points_to_mesh(points, res, threshold=None, denoise=True):
    """Generate mesh from point cloud.

    Briefly, the workflow is this:
      1. Partition the point cloud into voxels of size `res`.
      2. (Optional) Discard voxels with less than `threshold` points inside.
      3. Turn voxels into a (M, N, K) matrix.
      4. (Optional) Denoise the matrix by a round of binary erosion + dilation
         and fill holes.
      5. Use marching cubes to produce a mesh.

    Parameters
    ----------
    points :    (N, 3) array
                Point cloud.
    res :       int
                Resolution of the voxels.
    threshold : int, optional
                Use this to ignore voxels with very few points inside.
    denoise :   bool
                Whether to use binary filters to reduce noise and smoothen
                the mesh.


    Returns
    -------
    trimesh.Trimesh

    """
    if not skimage:
        raise ModuleNotFoundError(
            'Meshing requires `skimage`:\n'
            '  pip3 install scikit-image'
            )

    points = np.asarray(points)

    if points.ndim != 2 or points.shape[1] != 3:
        raise ValueError(f'Points must be of shape (N, 3), got {points.shape}')

    # Generate counts per voxel
    vxl, cnt = np.unique((points / res).round().astype(int),
                         return_counts=True, axis=0)

    # Turn into a DataFrame
    voxels = pd.DataFrame(np.vstack(vxl), columns=['x', 'y', 'z'])
    voxels['count'] = cnt

    if threshold:
        voxels = voxels[voxels['count'] > 1]

    # Generate empty matrix
    mat = np.zeros((voxels.x.max() + 1,
                    voxels.y.max() + 1,
                    voxels.z.max() + 1))

    # Fill matrix
    mat[voxels.x, voxels.y, voxels.z] = 1

    if denoise:
        # Denoise by a round of erosion...
        mat = ndimage.binary_erosion(mat)

        # ... followed by two rounds of dilation to smoothen things out...
        mat = ndimage.binary_dilation(mat, iterations=2)

        # ... followed by a round of fill holes
        mat = ndimage.binary_fill_holes(mat)

        # And a final round of erosion to get back to the correct scale
        mat = ndimage.binary_erosion(mat, iterations=1)

    # Run the marching cube algorithm
    # (newer versions of skimage have a "marching cubes" function and
    # the marching_cubes_lewiner is deprecreated)
    marching_cubes = getattr(measure, 'marching_cubes',
                             getattr(measure, 'marching_cubes_lewiner', None))
    verts, faces, normals, values = marching_cubes(mat.astype(float),
                                                   level=0,
                                                   gradient_direction='ascent',
                                                   allow_degenerate=False,
                                                   step_size=1)
    # Turn coordinates back into original units
    verts *= res

    # Somehow we seem to have introduced an offset equal to our resolution
    verts -= res

    mesh = tm.Trimesh(vertices=verts, faces=faces, normals=normals)

    # Need to fix normals
    mesh.fix_normals()

    return mesh


def pointlabels_to_meshes(points, labels, res, method='kde',
                          threshold=0.05, drop_fluff=True, volume=None,
                          n_cores=os.cpu_count() // 2, progress=True):
    """Generate non-overlapping meshes from a labelled point cloud.

    Briefly, the default workflow is this:

      1. Create a Gaussian KDE for each unique label.
      2. Tile the point's bounding box into voxels of size `res`.
      3. Calculate the KDE's point density function (PDF) to assign a label to
         each voxel.
      4. (Optional) Denoise the matrix by a round of binary erosion + dilation
         and fill holes.
      5. Use marching cubes to produce a mesh.

    See `method` parameter for an alternative method.

    Parameters
    ----------
    points :    (N, 3) array
                Point cloud.
    labels :    (N, ) array
                A label for each point.
    res :       int
                Size of the voxels. Note that this also determines the
                gap between meshes: higher resolution = smaller, more precise gaps.
    method :    "kde" | "majority"
                Which method to use. "kde" (default) uses above described
                workflow. "majority" vote will simply ask, for each voxel, which
                labels are contained within it and which is the most numerous.
                The latter is much faster but may produce coarser meshes - in
                particular if the number of points is low.
    threshold : float [0-1], optional
                Threshold for dropping voxels that don't appear to belong to
                any of the original labels. This is the quantile! I.e. the
                default value of 0.05 means that we'll be dropping the bottom 5%.
    drop_fluff : bool
                Whether to drop small bits and pieces from meshes and only keep
                the largest contiguous pieces.
    volume :    Volume | Trimesh, optional
                Provide a mesh to contrain the sampled voxels to inside this
                volume.
    n_cores :   int
                Number of cores to use for parallel processing. Each unique
                label will be processed on a separate core.


    Returns
    -------
    meshes  :   list
                List of `navis.Volume`. Their names correspond to unique
                `labels`.

    """
    if not skimage:
        raise ModuleNotFoundError(
            'Meshing requires `skimage`:\n'
            '   pip3 install scikit-image'
            )

    if len(points) != len(labels):
        raise ValueError(f'Number of labels ({len(labels)}) must match number '
                         f'of points ({len(points)})')

    assert method in ('kde', 'majority')

    points = np.asarray(points)
    labels = np.asarray(labels)
    labels_unique = np.unique(labels)

    if method == 'kde':
        # Now create voxel coordinates for the volume we want to fill:
        # First the bounding box
        bbox = np.vstack((points.min(axis=0), points.max(axis=0)))

        # We'll pad the volume by 2x the resolution
        padding = res * 2

        xco = np.arange(bbox[0][0] - padding, bbox[1][0] + padding, res)
        yco = np.arange(bbox[0][1] - padding, bbox[1][1] + padding, res)
        zco = np.arange(bbox[0][2] - padding, bbox[1][2] + padding, res)

        i = np.arange(0, xco.shape[0], 1)
        j = np.arange(0, yco.shape[0], 1)
        k = np.arange(0, zco.shape[0], 1)

        ii, jj, kk = np.meshgrid(i, j, k)
        xx, yy, zz = np.meshgrid(xco, yco, zco)

        voxels = np.vstack((xx.flatten(), yy.flatten(), zz.flatten())).T
        voxels = pd.DataFrame(voxels, columns=['x', 'y', 'z'])

        voxels['i'] = ii.flatten()
        voxels['j'] = jj.flatten()
        voxels['k'] = kk.flatten()

        if not isinstance(volume, type(None)):
            in_vol = intersection.in_volume(voxels[['x', 'y', 'z']].values,
                                            volume)
            print(f'Dropping {(~in_vol).sum()}/{len(voxels)} voxels outside of provided volume.')
            voxels = voxels.loc[in_vol].copy()

        # For each label create a KDE
        kde = {}
        for l in tqdm(labels_unique,
                    desc='Generating KDEs',
                    disable=not progress,
                    leave=False):
            this_p = points[labels == l]

            kde[l] = stats.gaussian_kde(this_p.T)

        # For each point get the point density function for each KDE
        combinations = [(kde[l], [voxels[['x', 'y', 'z']].values.T], {}) for l in kde]
        with mp.Pool(n_cores) as pool:
            results = list(tqdm(pool.imap(_worker_wrapper,
                                        combinations,
                                        chunksize=1),
                                leave=False,
                                disable=not progress,
                                total=len(combinations),
                                desc=f'Assigning {len(voxels):,} voxels'))

        # Fill results
        for l, r in zip(kde, results):
            voxels[l] = r

        # Drop voxels that have a PDF of less than the given threshold
        if threshold:
            pdf = voxels[labels_unique].max(axis=1)
            keep = pdf >= np.quantile(pdf, threshold)
            print(f'Dropping {(~keep).sum()}/{len(voxels)} voxels with too low density.')
            voxels = voxels.loc[keep].copy()

        # Assign label to each voxel based on the max probability
        voxels['label'] = labels_unique[np.argmax(voxels[labels_unique].values, axis=1)]
    else:
        # Turn points into voxels of given size
        points_vxl = points // res

        # Turn labels into an array of integers
        labels_dict = dict(zip(labels_unique, np.arange(len(labels_unique))))
        labels_int = np.array([labels_dict[la] for la in labels])

        # Combine (N, 3) voxels and (N, ) labels into (N, 4) array
        point_labels = np.hstack((points_vxl, labels_int.reshape(-1, 1)))

        # Count unique voxel + label combinations
        point_labels_unique, cnt = np.unique(point_labels, axis=0, return_counts=True)

        # For each x/y/z coordinate (but not the label column) get a unique index
        unique_voxel, voxel_ix = np.unique(
            point_labels_unique[:, :3], axis=0, return_inverse=True
        )

        # Turn into DataFrame for tallying up
        point_labels_df = pd.DataFrame()
        point_labels_df["voxel_ix"] = voxel_ix
        point_labels_df["label_ix"] = point_labels_unique[:, -1]
        point_labels_df["cnt"] = cnt

        # Turn into a N_point x N_labels matrix
        adj = point_labels_df.groupby(["voxel_ix", "label_ix"]).cnt.sum().unstack()

        # Now generate the DataFrame we will use to create meshes
        voxels = pd.DataFrame()
        voxels['i'] = unique_voxel[:, 0]
        voxels['j'] = unique_voxel[:, 1]
        voxels['k'] = unique_voxel[:, 2]

        # Also re-generate x/y/z coordinates
        voxels['x'] = voxels['i'] * res
        voxels['y'] = voxels['j'] * res
        voxels['z'] = voxels['k'] * res

        # Make sure i/j/k start at zero (otherwise matrix further down might
        # end up really huge)
        voxels['i'] -= unique_voxel[:, 0].min()
        voxels['j'] -= unique_voxel[:, 1].min()
        voxels['k'] -= unique_voxel[:, 2].min()

        # For each voxel get the top label...
        voxels['label'] = labels_unique[np.nanargmax(adj, axis=1)]

        # Drop voxels that have less than given points inside
        if threshold:
            top_count = np.nanmax(adj, axis=1)
            keep = top_count >= np.quantile(top_count, threshold)
            print(f'Dropping {(~keep).sum()}/{len(voxels)} voxels with too low density.')
            voxels = voxels.loc[keep].copy()

        # Some settings for the meshing
        padding = 0

    meshes = []
    for l in tqdm(labels_unique,
                  desc='Creating meshes',
                  disable=not progress,
                  leave=False):
        # Generate empty matrix
        mat = np.zeros((voxels.i.max() + 2,
                        voxels.j.max() + 2,
                        voxels.k.max() + 2))

        # Get voxels belonging to this label
        this = voxels[voxels.label == l]

        if this.empty:
            logger.warning(f'Label {l} did not produce a mesh.')
            continue

        # Fill matrix
        mat[this.i, this.j, this.k] = 1

        # Remove binary holes
        mat = ndimage.binary_fill_holes(mat)

        # We need one round of erodes to make meshes non-overlapping
        mat = ndimage.binary_erosion(mat)

        if not np.any(mat):
            logger.warning(f'Label {l} did not produce a mesh.')
            continue

        # Use marching cubes to create surface model
        # (newer versions of skimage have a "marching cubes" function and
        # the marching_cubes_lewiner is deprecreated)
        marching_cubes = getattr(measure, 'marching_cubes',
                                 getattr(measure, 'marching_cubes_lewiner', None))
        verts, faces, normals, values = marching_cubes(mat.astype(float),
                                                       level=0,
                                                       allow_degenerate=False,
                                                       step_size=1)

        # Scale back to original units
        verts *= res

        # Add offset
        offset = voxels[['x', 'y', 'z']].min(axis=0).values  # keep .values !
        verts += offset

        # Somehow we seem to have introduced an offset
        verts -= padding
        verts += 0.5 * res

        if method == 'kde':
            verts[:, 1] -= 0.5 * res

        # Make a trimesh
        new_mesh = tm.Trimesh(vertices=verts, faces=faces, normals=normals)

        if drop_fluff:
            # Drop small stuff (anything that makes up less than 10% of the faces)
            cc = tm.graph.connected_components(edges=new_mesh.face_adjacency,
                                               nodes=np.arange(len(new_mesh.faces)),
                                               min_len=1,
                                               engine=None)
            if len(cc) > 1:
                min_faces = new_mesh.faces.shape[0] * 0.1
                to_keep = [c for c in cc if (len(c) >= min_faces)]
                if to_keep:
                    new_mesh = new_mesh.submesh([np.concatenate(to_keep)])[0]

        # Need to fix normals
        new_mesh.fix_normals()

        meshes.append(core.Volume(new_mesh, name=l))

    return meshes


def _worker_wrapper(x):
    f, args, kwargs = x
    return f(*args, **kwargs)



def face_dist_sorting(x, from_, strahler_weight=False, inplace=False):
    """Sort faces by distance from given point.

    This allows you to e.g. use Blender's "build" modifier to grow neurons
    from a point of origin.

    Parameters
    ----------
    x :         navis.MeshNeuron
                Mesh to sort faces for.
    from_ :     int | list of int
                Must be either a vertex index (single int) or an x/y/z coordinate.
    strahler_weight :   bool
                If True, will use Strahler index to grow twigs slower than
                backbone.
    inplace :   bool
                Whether to modify the input mesh or a copy thereof.

    Returns
    -------
    navis.MeshNeuron

    Examples
    --------

    >>> import navis
    >>> x = navis.example_neurons(1, kind='mesh')
    >>> x = navis.meshes.mesh_utils.face_dist_sorting(x, from_=x.soma_pos)

    """
    # Turn vertex indices to coordinates
    if isinstance(from_, (int, np.integer)):
        from_ = x.vertices[from_]

    # Generate the skeleton
    # (note we don't shave to avoid issues with vertex map)
    sk = x.skeletonize(heal=True, shave=False)

    # Get the node index for our from_
    seed = sk.snap(from_)[0]

    # Get distances from
    dists = graph.geodesic_matrix(sk, from_=seed).iloc[0]

    if strahler_weight:
        sk.reroot(seed, inplace=True)
        _ = morpho.strahler_index(sk)
        dists = dists / sk.nodes.strahler_index

    # Get sorting by distance
    srt = np.argsort(dists.values)

    # Map sorting back onto vertices
    verts_srt = srt[sk.vertex_map]

    # For each face get the skeleton nodes it maps to
    face_nodes = sk.vertex_map[x.faces]

    # For each face get the mean distance
    face_dist = dists.values[face_nodes].mean(axis=1)

    # Sort the faces
    faces_srt = np.array(x.faces)[np.argsort(face_dist)]

    if not inplace:
        x = x.copy()

    x.faces = faces_srt

    return x



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import os
import json

import networkx as nx
import pandas as pd

from typing import Union, Optional
from typing_extensions import Literal

from ..core import Volume, NeuronList, TreeNeuron, MeshNeuron
from ..io import read_swc
from ..graph import nx2neuron

__all__ = ['example_neurons', 'example_volume']

fp = os.path.dirname(__file__)

gml_path = os.path.join(fp, 'gml')
swc_path = os.path.join(fp, 'swc')
vols_path = os.path.join(fp, 'volumes')
obj_path = os.path.join(fp, 'obj')
syn_path = os.path.join(fp, 'synapses')

gml = sorted([f for f in os.listdir(gml_path) if f.endswith('.gml')])
swc = sorted([f for f in os.listdir(swc_path) if f.endswith('.swc')])
vols = sorted([f for f in os.listdir(vols_path) if f.endswith('.obj')])
obj = sorted([f for f in os.listdir(obj_path) if f.endswith('.obj')])
syn = sorted([f for f in os.listdir(syn_path) if f.endswith('.csv')])

NeuronObject = Union[TreeNeuron, MeshNeuron, NeuronList]

# Soma positions for the example neurons (for the meshes)
SOMA_POS = {1734350788: [14957.1, 36540.7, 28432.4],
            1734350908: [15503.5, 35903.1, 23151.6],
            722817260: None,
            754534424: [15150, 35262.7, 23136.6],
            754538881: [13810, 35236, 25222.8]}


def example_neurons(n: Optional[int] = None,
                    kind:  Union[Literal['mesh'],
                                 Literal['skeleton'],
                                 Literal['mix']] = 'skeleton',
                    synapses: bool = True,
                    source: Union[Literal['swc'],
                                  Literal['gml']] = 'swc',
                    ) -> NeuronObject:
    """Load example neuron(s).

    These example neurons are skeletons and meshes of the same olfactory
    projection neurons from the DA1 glomerulus which have been automatically
    segmented in the Janelia hemibrain data set [1]. See also
    `https://neuprint.janelia.org`.

    Coordinates are in voxels which equal 8 x 8 x 8 nanometers.

    Parameters
    ----------
    n :         int | None, optional
                Number of neurons to return. If None, will return all available
                example neurons. Can never return more than the maximum number
                of available example neurons.
    kind :      "skeleton" | "mesh" | "mix"
                What kind of neurons to return.
    synapses :  bool,
                If True, will also load synapses.
    source :    'swc' | 'gml', optional
                Only relevant for skeletons. Skeletons can be generated from SWC
                files or GML graphs (this is really only used for testing).

    Returns
    -------
    TreeNeuron
                If `n=1` and `kind='skeleton'`.
    MeshNeuron
                If `n=1` and `kind='mesh'`.
    NeuronList
                List of the above neuron types if `n>1`.

    References
    ----------
    [1] Louis K. Scheffer et al., bioRxiv. 2020. doi: https://doi.org/10.1101/2020.04.07.030213
    A Connectome and Analysis of the Adult Drosophila Central Brain.

    Examples
    --------
    Load a single neuron

    >>> import navis
    >>> n = navis.example_neurons(n=1)

    Load all example neurons

    >>> nl = navis.example_neurons()

    """
    if kind not in ['skeleton', 'mesh', 'mix']:
        raise ValueError(f'Unknown value for `kind`: "{kind}"')

    if isinstance(n, type(None)):
        if kind == 'mix':
            n = len(swc) + len(obj)
        else:
            n = len(swc)
    elif not isinstance(n, int):
        raise TypeError(f'Expected int or None, got "{type(n)}"')

    if isinstance(n, int) and n < 1:
        raise ValueError("Unable to return less than 1 neuron.")

    if kind == 'mix':
        n_mesh = round(n/2)
        n_skel = n - n_mesh
    else:
        n_mesh = n_skel = n

    nl = []
    if kind in ['skeleton', 'mix']:
        if source == 'gml':
            graphs = [nx.read_gml(os.path.join(gml_path, g)) for g in gml[:n_skel]]
            nl += [nx2neuron(g,
                             units='8 nm',
                             id=int(f.split('.')[0])) for f, g in zip(gml, graphs)]
        elif source == 'swc':
            nl += [read_swc(os.path.join(swc_path, f),
                            units='8 nm',
                            id=int(f.split('.')[0])) for f in swc[:n_skel]]
        else:
            raise ValueError(f'Source must be "swc" or "gml", not "{source}"')

    if kind in ['mesh', 'mix']:
        files = [os.path.join(obj_path, f) for f in obj[:n_mesh]]
        nl += [MeshNeuron(fp,
                          units='8 nm',
                          name=f.split('.')[0],
                          id=int(f.split('.')[0])) for f, fp in zip(obj, files)]
        for n in nl:
            n.soma_pos = SOMA_POS[n.id]

    if synapses:
        for n in nl:
            n.connectors = pd.read_csv(os.path.join(syn_path, f'{n.id}.csv'))

            if isinstance(n, MeshNeuron):
                n._connectors.drop('node_id', axis=1, inplace=True)

    with open(os.path.join(fp, 'meta.json'), 'r') as f:
        meta = json.load(f)

    for n in nl:
        n.name = meta[str(n.id)]['instance']

    if len(nl) == 1:
        return nl[0]
    return NeuronList(nl)


def example_volume(name: str) -> Volume:
    """Load an example volume.

    Volumes are in hemibrain space which means coordinates are in voxels
    at 8 x 8 x 8 nanometers/voxel.

    Parameters
    ----------
    name :      str
                Name of available volume. Currently available::

                  "LH" = lateral horn in hemibrain space
                  "neuropil" = neuropil in hemibrain space

    Returns
    -------
    navis.Volume

    Examples
    --------
    Load LH volume

    >>> import navis
    >>> lh = navis.example_volume('LH')

    """
    if not isinstance(name, str):
        raise TypeError(f'Expected string, got "{type(name)}"')

    # Force lower case
    name = name.lower()

    # Make sure extension is correct
    if not name.endswith(".obj"):
        name += ".obj"

    if name not in vols:
        raise ValueError(
            f'No volume named "{name}". Available volumes: {",".join(vols)}'
        )

    vol = Volume.from_file(
        os.path.join(vols_path, name), name=name.split(".")[0], units="nm"
    )

    return vol



import navis
import pytest
import tempfile
import numpy as np

from pathlib import Path


@pytest.mark.parametrize("filename", ['', '{neuron.id}.swc',
                                      'neurons.zip',
                                      '{neuron.id}.swc@neurons.zip'])
def test_swc_io(filename):
    with tempfile.TemporaryDirectory() as tempdir:
        filepath = Path(tempdir) / filename

        # Load example neurons
        n = navis.example_neurons(2, kind='skeleton')

        # Save to file / folder
        navis.write_swc(n, filepath)

        # Load again
        if str(filepath).endswith('.zip'):
            n2 = navis.read_swc(Path(tempdir) / 'neurons.zip')
        else:
            n2 = navis.read_swc(tempdir)

        # Assert that we loaded the same number of neurons
        assert len(n) == len(n2)


@pytest.mark.parametrize("filename", ['',
                                      'neurons.zip',
                                      '{neuron.id}@neurons.zip'])
def test_precomputed_skeleton_io(filename):
    with tempfile.TemporaryDirectory() as tempdir:
        filepath = Path(tempdir) / filename

        # Load example neurons
        n = navis.example_neurons(2, kind='skeleton')

        # Save to file / folder
        navis.write_precomputed(n, filepath, radius=True)

        # Load again
        if str(filepath).endswith('.zip'):
            n2 = navis.read_precomputed(Path(tempdir) / 'neurons.zip')
        else:
            n2 = navis.read_precomputed(tempdir)

        # Assert that we loaded the same number of neurons
        assert len(n) == len(n2)


@pytest.mark.parametrize("filename", ['',
                                      'neurons.zip',
                                      '{neuron.id}@neurons.zip'])
def test_precomputed_mesh_io(filename):
    with tempfile.TemporaryDirectory() as tempdir:
        filepath = Path(tempdir) / filename

        # Load example neurons
        n = navis.example_neurons(2, kind='mesh')

        # Save to file / folder
        navis.write_precomputed(n, filepath, write_manifest=True)

        # Load again
        if str(filepath).endswith('.zip'):
            n2 = navis.read_precomputed(Path(tempdir) / 'neurons.zip')
        else:
            n2 = navis.read_precomputed(tempdir)

        # Assert that we loaded the same number of neurons
        assert len(n) == len(n2)


@pytest.mark.parametrize("filename", ['neurons.zip',
                                      '*.ply'])
def test_mesh_io(filename):
    with tempfile.TemporaryDirectory() as tempdir:
        tempdir = Path(tempdir)

        # Load example neurons
        n = navis.example_neurons(2, kind='mesh')

        # Save to neurons folder
        if str(filename).endswith('.zip'):
            # Into a zip file
            navis.write_mesh(n, tempdir / 'neurons.zip', filetype='ply')
        else:
            # As individual files
            navis.write_mesh(n, tempdir, filetype='ply')

        # Load again
        if str(filename).endswith('.zip'):
            n2 = navis.read_mesh(tempdir / 'neurons.zip')
        else:
            n2 = navis.read_mesh(tempdir / filename)

        # Assert that we loaded the same number of neurons
        assert len(n) == len(n2)


def test_read_nrrd(voxel_nrrd_path):
    navis.read_nrrd(voxel_nrrd_path, output="voxels", errors="raise")


def test_roundtrip_nrrd(voxel_nrrd_path):
    vneuron = navis.read_nrrd(voxel_nrrd_path, output="voxels", errors="raise")
    outpath = voxel_nrrd_path.parent / "written.nrrd"
    navis.write_nrrd(vneuron, outpath)
    vneuron2 = navis.read_nrrd(outpath, output="voxels", errors="raise")
    assert np.allclose(vneuron._data, vneuron2._data)
    assert np.allclose(vneuron.units_xyz.magnitude, vneuron2.units_xyz.magnitude)
    assert vneuron.units_xyz.units == vneuron2.units_xyz.units
